# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
import time

from ._logging import LogExtraInfo, LogContent
from psutil import NoSuchProcess, Process, wait_procs, STATUS_STOPPED
from typing import List


def _suspend_process(logger, process: Process) -> bool:
    """
    Suspend a given process.

    Parameters:
    - logger: The logging instance for logging.
    - process: The process to be suspended.

    Returns:
    - True if the process was successfully suspended, False otherwise.
    """
    try:
        process.suspend()
        for _ in range(10):
            if process.status() == STATUS_STOPPED:
                return True
            # Wait for the process to be suspended
            time.sleep(0.1)
    except Exception as e:
        logger.error(
            f"Failed to suspend process {process.pid}: {e}",
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )
        return False

    return False


def _suspend_process_tree(
    logger,
    process: Process,
    all_processes: List[Process],
    procs_cannot_suspend: List[Process],
    suspend_subprocesses: bool,
) -> None:
    """
    Recursively suspend the process tree and its children in pre-order.

    Parameters:
    - logger: The logging instance for logging.
    - process: The process to start suspension from.
    - all_processes: List of all processes we are trying to suspend.
    - procs_cannot_suspend: List of processes that couldn't be suspended.
    - suspend_subprocesses: Control if the child processes needed to be suspended
    """
    # Attempt to suspend the current process.
    if not _suspend_process(logger, process):
        procs_cannot_suspend.append(process)

    all_processes.append(process)

    if not suspend_subprocesses:
        return

    # Recursively suspend child processes.
    for child in process.children():
        _suspend_process_tree(
            logger, child, all_processes, procs_cannot_suspend, suspend_subprocesses
        )


def _kill_processes(logger, process_list: List[Process]) -> List[Process]:
    """
    Kill all processes in the given list.

    Parameters:
    - logger: The logging instance for logging.
    - process_list: List of processes to be killed.

    Returns:
    - List of processes that are still alive after attempting to kill.
    """

    for process in process_list:
        try:
            logger.info(
                f"Killing process with id {process.pid}.",
                extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
            )
            process.kill()
        except NoSuchProcess:
            logger.info(
                f"No process with id {process.pid} for termination",
                extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
            )
        except Exception as e:
            logger.error(
                f"Failed to kill process {process.pid}: {e}",
                extra=LogExtraInfo(
                    openjd_log_content=LogContent.PROCESS_CONTROL | LogContent.EXCEPTION_INFO
                ),
            )

    # Wait for the processes to be terminated
    _, alive = wait_procs(process_list, timeout=5)
    return alive


def kill_windows_process_tree(logger, root_pid, signal_subprocesses=True) -> None:
    """
    Kills the windows process tree starting from the provided root PID.
    The termination ordering will be from leafs to root.

    Parameters:
    - logger: The logging instance to record activities.
    - root_pid: The root process ID.
    - signal_subprocesses: Flag to determine if subprocesses should be signaled.
    """

    try:
        parent_process = Process(root_pid)
    except NoSuchProcess:
        logger.error(
            f"Root process with PID {root_pid} not found.",
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )
        return

    procs_failed_to_suspend: List[Process] = []
    processes_to_be_killed: List[Process] = []
    _suspend_process_tree(
        logger, parent_process, processes_to_be_killed, procs_failed_to_suspend, signal_subprocesses
    )
    if procs_failed_to_suspend:
        logger.warning(
            f"Following processes cannot be suspended. Processes IDs: {[proc.pid for proc in procs_failed_to_suspend]}",
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )

    # Ensure we kill child processes first
    processes_to_be_killed.reverse()
    alive_processes = _kill_processes(logger, processes_to_be_killed)

    if alive_processes:
        logger.warning(
            f"Failed to kill following process(es): {[p.pid for p in alive_processes]}. Retrying...",
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )
        alive_processes = _kill_processes(logger, processes_to_be_killed)
        if alive_processes:
            logger.warning(
                f"Still failed to kill the following process(es): {[p.pid for p in alive_processes]}. Please handle manually.",
                extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
            )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from __future__ import annotations

import json
import logging
import re
from enum import Enum
from typing import Any, Callable

from ._logging import LOG, LogContent, LogExtraInfo

__all__ = ("ActionMessageKind", "ActionMonitoringFilter")


class ActionMessageKind(Enum):
    PROGRESS = "progress"  # A progress percentile for the running action
    STATUS = "status"  # A status message
    FAIL = "fail"  # A failure message
    ENV = "env"  # Defining an environment variable
    UNSET_ENV = "unset_env"  # Unsetting an environment variable

    # The following are not in the spec, but are utility provided by this runtime.
    SESSION_RUNTIME_LOGLEVEL = "session_runtime_loglevel"  # Setting the log level of this runtime


# A composite regex that matches one of the message kinds to a named capture group
# with the same name as the message kind.
filter_regex = (
    "^openjd_(?:"
    f"{'|'.join(f'(?P<{re.escape(v.value)}>{re.escape(v.value)})' for v in ActionMessageKind)}"
    "): (.+)$"
)
filter_matcher = re.compile(filter_regex)

openjd_env_actions_filter_regex = "^(openjd_env|openjd_unset_env)"
openjd_env_actions_filter_matcher = re.compile(openjd_env_actions_filter_regex)

# A regex for matching the assignment of a value to an environment variable
envvar_set_regex_str = "^[A-Za-z_][A-Za-z0-9_]*" "=" ".*$"  # Variable name
envvar_set_regex_json = '^(")?[A-Za-z_][A-Za-z0-9_]*' "=" ".*$"  # Variable name
envvar_set_matcher_str = re.compile(envvar_set_regex_str)
envvar_set_matcher_json = re.compile(envvar_set_regex_json)
envvar_unset_regex = "^[A-Za-z_][A-Za-z0-9_]*$"
envvar_unset_matcher = re.compile(envvar_unset_regex)


# This is a reworking/merging of TaskStatusFilter and FailureFilter
class ActionMonitoringFilter(logging.Filter):
    """Captures any Open Job Description-defined updates from the subprocess that are communicated
    in the form of single lines in stdout of the form:
    openjd_progress: <progress in the form of a float between 0.0 and 100.0>
    openjd_status: <string indicating the new status>
    openjd_fail: <string indicating a failure message>
    openjd_env: <env var name>=<string value>
    openjd_unset_env: <env var name>
    openjd_session_runtime_loglevel: [ERROR | WARNING | INFO | DEBUG]

    When such a message is detected in the log stream a given callback will be
    called with the details of the update message. The callback will be called
    with arguments:
        callback(ActionMessageKind.PROGRESS, <float between 0.0 and 100.0>)
        callback(ActionMessageKind.STATUS, <string indicating the new status>)
        callback(ActionMessageKind.FAIL, <string indicating a failure message>)
        callback(ActionMessageKind.ENV, {"name": <envvar name>, "value": <envvar value>})
        callback(ActionMessageKind.UNSET_ENV, <string indicating the name of the env var>)
        callback(ActionMessageKind.RUNTIME_LOGLEVEL, <integer log level>)
    """

    _session_id: str
    """The id that we're looking for in LogRecords.
    We only process records with the "session_id" attribute set to this value.
    """

    _callback: Callable[[ActionMessageKind, Any, bool], None]
    """Callback to invoke when one of the Open Job Description update messages is detected.
    Args:
        [0]: The kind of the update message.
        [1]: The information/message given after the Open Job Description message prefix ("openjd_<name>: ")
        [2]: A boolean to express whether or not the corresponding Action is to be Canceled & marked Failed
    """

    _suppress_filtered: bool
    """If true, then any Open Job Description output stream messages are removed from the log
    when filtering."""

    _internal_handlers: dict[ActionMessageKind, Callable[[str], None]]
    """A mapping from message kind to the specfic ActionMonitoringFilter method that
    will handle processing the message type."""

    # The range of allowable values for progress reporting
    _MIN_PROGRESS: float = 0.0
    _MAX_PROGRESS: float = 100.0

    def __init__(
        self,
        name: str = "",
        *,
        session_id: str,
        callback: Callable[[ActionMessageKind, Any, bool], None],
        suppress_filtered: bool = False,
    ):
        """
        Args:
            name (str, optional): If name is specified, it names a logger which, together
                with its children, will have its events allowed through the filter. If name
                is the empty string, allows every event. Defaults to "".
            session_id (str): The id that we're looking for in LogRecords.
                We only process records with the "session_id" attribute set to this value.
            callback (Callable[[ActionMessageKind,Any], None]): A callable to call
                with a message payload when an Open Job Description message is found in the log.
            suppress_filtered (bool, optional): If True, then all Open Job Description messages
                will be filtered out of the log. Defaults to True.
        """
        super().__init__(name)
        self._session_id = session_id
        self._callback = callback
        self._suppress_filtered = suppress_filtered
        self._internal_handlers = {
            ActionMessageKind.PROGRESS: self._handle_progress,
            ActionMessageKind.STATUS: self._handle_status,
            ActionMessageKind.FAIL: self._handle_fail,
            ActionMessageKind.ENV: self._handle_env,
            ActionMessageKind.UNSET_ENV: self._handle_unset_env,
            ActionMessageKind.SESSION_RUNTIME_LOGLEVEL: self._handle_session_runtime_loglevel,
        }

    def filter(self, record: logging.LogRecord) -> bool:
        """Called automatically by Python's logging subsystem when a log record
        is sent to a log to which this filter class is applied.

        If the LogRecord does not have a 'session_id' attribute, or if the value of
        the attribute differs from this filter's session_id then the filter does nothing.

        Args:
            record (logging.LogRecord): Log record that was sent to the log.

        Returns:
            bool: If true then the Python logger will keep the record in the log,
                  else it will remove it.
        """
        if not hasattr(record, "session_id") or getattr(record, "session_id") != self._session_id:
            # Not a record for us to process
            return True
        if not isinstance(record.msg, str):
            # If something sends a non-string to the logger (e.g. via logger.exception) then
            # don't try to string match it.
            return True
        match = filter_matcher.match(record.msg)
        if match and match.lastindex is not None:
            message = match.group(match.lastindex)
            # Note: keys of match.groupdict() are the names of named groups in the regex
            matched_named_groups = tuple(k for k, v in match.groupdict().items() if v is not None)
            if len(matched_named_groups) > 1:
                # The only way that this happens is if filter_matcher is constructed incorrectly.
                all_matched_groups = ",".join(k for k in matched_named_groups)
                LOG.error(
                    f"Open Job Description: Malformed output stream filter matched multiple kinds ({all_matched_groups})",
                    extra=LogExtraInfo(openjd_log_content=LogContent.COMMAND_OUTPUT),
                )
                return True
            message_kind = ActionMessageKind(matched_named_groups[0])
            try:
                handler = self._internal_handlers[message_kind]
            except KeyError:
                LOG.error(
                    f"Open Job Description: Unhandled message kind ({message_kind.value})",
                    extra=LogExtraInfo(openjd_log_content=LogContent.COMMAND_OUTPUT),
                )
                return True
            try:
                handler(message)
            except ValueError as e:
                record.msg = record.msg + f" -- ERROR: {str(e)}"
                # There was an error. Don't suppress the message from the log.
                return True
            return not self._suppress_filtered

        # Check for "almost" matching openjd_env and openjd_unset_env commands
        lower_case_trimmed_msg: str = record.msg.lstrip().lower()
        if openjd_env_actions_filter_matcher.match(lower_case_trimmed_msg):
            # There was a minor error like spaces or case in the env commands
            err_message = (
                f"Open Job Description: Incorrectly formatted openjd env command ({record.msg})"
            )
            record.msg = record.msg + f" -- ERROR: {err_message}"

            # Callback to cancel the action and mark it as FAILED
            self._callback(ActionMessageKind.FAIL, err_message, True)
            return True

        return True

    def _handle_progress(self, message: str) -> None:
        """Local handling of Progress messages. Processes the message and then
        calls the provided handler,

        Args:
            message (str): The message after the leading 'openjd_progress: ' prefix
        """

        try:
            progress = float(message)
            if not (self._MIN_PROGRESS <= progress <= self._MAX_PROGRESS):
                raise ValueError()
            self._callback(ActionMessageKind.PROGRESS, progress, False)
        except ValueError:
            raise ValueError(
                f"Progress must be a floating point value between {self._MIN_PROGRESS} and {self._MAX_PROGRESS}, inclusive."
            )

    def _handle_status(self, message: str) -> None:
        """Local handling of Status messages. Just passes the message directly to
        the callback.

        Args:
            message (str): The message after the leading 'openjd_status: ' prefix
        """
        self._callback(ActionMessageKind.STATUS, message, False)

    def _handle_fail(self, message: str) -> None:
        """Local handling of Fail messages. Just passes the message directly to
        the callback.

        Args:
            message (str): The message after the leading 'openjd_fail: ' prefix
        """
        self._callback(ActionMessageKind.FAIL, message, False)

    def _handle_env(self, message: str) -> None:
        """Local handling of the Env messages.

        Args:
            message (str): The message after the leading 'openjd_env: ' prefix
        """
        message = message.lstrip()
        # A correctly formed message is of the form:
        # <varname>=<value>
        # where:
        #   <varname> consists of latin alphanumeric characters and the underscore,
        #             and starts with a non-digit
        #   <value> can be any characters including empty.
        if not envvar_set_matcher_str.match(message) and not envvar_set_matcher_json.match(message):
            err_message = "Failed to parse environment variable assignment."
            # Callback to fail and cancel action on this error
            self._callback(ActionMessageKind.ENV, err_message, True)
            raise ValueError(err_message)
        elif envvar_set_matcher_str.match(message):
            name, _, value = message.partition("=")
        else:
            message_json_str = json.loads(message)
            name, _, value = message_json_str.partition("=")

        self._callback(ActionMessageKind.ENV, {"name": name, "value": value}, False)

    def _handle_unset_env(self, message: str) -> None:
        """Local handling of the unset env messages.

        Args:
            message (str): The message after the leading 'openjd_unset_env: ' prefix
        """
        message = message.lstrip()
        # A correctly formed message is of the form:
        # <varname>
        # where:
        #   <varname> consists of latin alphanumeric characters and the underscore,
        #             and starts with a non-digit
        if not envvar_unset_matcher.match(message):
            err_message = "Failed to parse environment variable name."
            # Callback to fail and cancel action on this error
            self._callback(ActionMessageKind.UNSET_ENV, err_message, True)
            raise ValueError(err_message)
        self._callback(ActionMessageKind.UNSET_ENV, message, False)

    def _handle_session_runtime_loglevel(self, message: str) -> None:
        """Local handling of the session runtime loglevel messages.

        Args:
            message (str): The message after the leading 'openjd_session_runtime_loglevel: ' prefix
        """
        message = message.upper().strip()
        levels = {
            "DEBUG": logging.DEBUG,
            "INFO": logging.INFO,
            "WARNING": logging.WARNING,
            "ERROR": logging.ERROR,
        }
        loglevel = levels.get(message, None)
        if loglevel is not None:
            self._callback(ActionMessageKind.SESSION_RUNTIME_LOGLEVEL, loglevel, False)
        else:
            raise ValueError(
                f"Unknown log level: {message}. Known values: {','.join(levels.keys())}"
            )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import sys
import ctypes
from ctypes.wintypes import (
    BOOL,
    DWORD,
    HANDLE,
    LONG,
    LPCWSTR,
    LPDWORD,
    LPVOID,
    LPWSTR,
    PBYTE,
    PDWORD,
    PHANDLE,
    ULONG,
    WORD,
)
from ctypes import POINTER, WinError, byref, c_byte, c_size_t, c_void_p, pointer  # type: ignore
from collections.abc import Sequence

# This assertion short-circuits mypy from type checking this module on platforms other than Windows
# https://mypy.readthedocs.io/en/stable/common_issues.html#python-version-and-system-platform-checks
assert sys.platform == "win32"

# =======================
# Constants
# =======================

# Ref: https://learn.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-createprocesswithtokenw
LOGON_WITH_PROFILE = 0x00000001
LOGON_NETCREDENTIALS_ONLY = 0x00000002

# STARTINFO Flags (ref: https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/ns-processthreadsapi-startupinfoa)
STARTF_USESHOWWINDOW = 0x00000001
STARTF_USESTDHANDLES = 0x00000100

# ShowWindow flags (ref: https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-showwindow)
SW_HIDE = 0

# TOKEN_PRIVILEGE attributes (ref: https://learn.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-token_privileges)
SE_PRIVILEGE_ENABLED_BY_DEFAULT = 0x00000001
SE_PRIVILEGE_ENABLED = 0x00000002
SE_PRIVILEGE_REMOVED = 0x00000004
SE_PRIVILEGE_USED_FOR_ACCESS = 0x80000000

# Ref: https://learn.microsoft.com/en-us/windows/win32/secauthz/privilege-constants
SE_BACKUP_NAME = "SeBackupPrivilege"
SE_RESTORE_NAME = "SeRestorePrivilege"
SE_INCREASE_QUOTA_NAME = "SeIncreaseQuotaPrivilege"
SE_ASSIGNPRIMARYTOKEN_NAME = "SeAssignPrimaryTokenPrivilege"
SE_TCB_NAME = "SeTcbPrivilege"

# Constant values (ref: https://learn.microsoft.com/en-us/windows/win32/secauthn/logonuserexexw)
LOGON32_PROVIDER_DEFAULT = 0
LOGON32_LOGON_INTERACTIVE = 2
LOGON32_LOGON_NETWORK = 3
LOGON32_LOGON_BATCH = 4
LOGON32_LOGON_SERVICE = 5
LOGON32_LOGON_NETWORK_CLEARTEXT = 8

# Prevents displaying of messages
PI_NOUI = 0x00000001

# Values of the TOKEN_TYPE enumeration
# https://learn.microsoft.com/en-us/windows/win32/api/winnt/ne-winnt-token_type
TOKEN_TYPE_PRIMARY = 1
TOKEN_TYPE_IMPERSONATION = 2

# Values of the SECURITY_IMPERSONATION_LEVEL enumeration (SIL = "Security Impersonation Level")
# https://learn.microsoft.com/en-us/windows/win32/api/winnt/ne-winnt-security_impersonation_level
SIL_ANONYMOUS = 0
SIL_IDENTIFICATION = 1
SIL_IMPERSONATION = 2
SIL_DELEGATION = 3

STANDARD_RIGHTS_REQUIRED = 0x0F0000
STANDARD_RIGHTS_READ = 0x020000
STANDARD_RIGHTS_WRITE = STANDARD_RIGHTS_READ
STANDARD_RIGHTS_EXECUTE = STANDARD_RIGHTS_READ
STANDARD_RIGHTS_ALL = 0x1F0000

# Token access privileges (ref: https://learn.microsoft.com/en-us/windows/win32/secauthz/access-rights-for-access-token-objects)
TOKEN_ASSIGN_PRIMARY = 0x0001
TOKEN_DUPLICATE = 0x0002
TOKEN_IMPERSONATE = 0x0004
TOKEN_QUERY = 0x0008
TOKEN_QUERY_SOURCE = 0x0010
TOKEN_ADJUST_PRIVILEGES = 0x0020
TOKEN_ADJUST_GROUPS = 0x0040
TOKEN_ADJUST_DEFAULT = 0x0080
TOKEN_ADJUST_SESSIONID = 0x0100
TOKEN_READ = STANDARD_RIGHTS_READ | TOKEN_QUERY
TOKEN_WRITE = (
    STANDARD_RIGHTS_WRITE | TOKEN_ADJUST_PRIVILEGES | TOKEN_ADJUST_GROUPS | TOKEN_ADJUST_DEFAULT
)
TOKEN_ALL_ACCESS = (
    STANDARD_RIGHTS_ALL
    | TOKEN_ASSIGN_PRIMARY
    | TOKEN_DUPLICATE
    | TOKEN_IMPERSONATE
    | TOKEN_QUERY
    | TOKEN_QUERY_SOURCE
    | TOKEN_ADJUST_PRIVILEGES
    | TOKEN_ADJUST_GROUPS
    | TOKEN_ADJUST_DEFAULT
    | TOKEN_ADJUST_SESSIONID
)

# From https://learn.microsoft.com/en-us/windows/win32/api/winnt/ne-winnt-token_information_class
TokenPrivileges = 3
TokenSecurityAttributes = 39

PROC_THREAD_ATTRIBUTE_HANDLE_LIST = 0x00020002

# =======================
# Structures/Types
# =======================

SIZE_T = c_size_t
PSIZE_T = POINTER(SIZE_T)


# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/ns-processthreadsapi-startupinfow
class STARTUPINFO(ctypes.Structure):
    _fields_ = [
        ("cb", DWORD),
        ("lpReserved", LPWSTR),
        ("lpDesktop", LPWSTR),
        ("lpTitle", LPWSTR),
        ("dwX", DWORD),
        ("dwY", DWORD),
        ("dwXSize", DWORD),
        ("dwYSize", DWORD),
        ("dwXCountChars", DWORD),
        ("dwYCountChars", DWORD),
        ("dwFillAttribute", DWORD),
        ("dwFlags", DWORD),
        ("wShowWindow", WORD),
        ("cbReserved2", WORD),
        ("lpReserved2", PBYTE),
        ("hStdInput", HANDLE),
        ("hStdOutput", HANDLE),
        ("hStdError", HANDLE),
    ]


PPROC_THREAD_ATTRIBUTE_LIST = c_void_p
LPPROC_THREAD_ATTRIBUTE_LIST = PPROC_THREAD_ATTRIBUTE_LIST


# https://learn.microsoft.com/en-us/windows/win32/api/winbase/ns-winbase-startupinfoexw
class STARTUPINFOEX(ctypes.Structure):
    _fields_ = [("StartupInfo", STARTUPINFO), ("lpAttributeList", PPROC_THREAD_ATTRIBUTE_LIST)]

    def allocate_attribute_list(self, num_attributes: int) -> None:
        """Allocate a buffer to lpAttributeList such that it can hold information for
        'num_attributes' attributes.
        Note: You must call 'deallocate_attribute_list()' when done with this structure if you call this;
           that will ensure that the additional allocations that the OS has made are deallocated.
        """
        # As per https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-initializeprocthreadattributelist#remarks
        # First we call InitializeProcThreadAttributeList with an null attribute list,
        # and it'll tell us how large of a buffer lpAttributeList needs to be.
        # This will always return False, so we don't check return code.
        lp_size = SIZE_T(0)
        InitializeProcThreadAttributeList(
            None, num_attributes, 0, byref(lp_size)  # reserved, and must be 0
        )

        # Allocate the desired buffer
        buffer = (c_byte * lp_size.value)()
        self.lpAttributeList = ctypes.cast(pointer(buffer), c_void_p)

        # Second call to actually initialize the buffer
        if not InitializeProcThreadAttributeList(
            self.lpAttributeList, num_attributes, 0, byref(lp_size)  # reserved, and must be 0
        ):
            raise WinError()

    def deallocate_attribute_list(self) -> None:
        DeleteProcThreadAttributeList(self.lpAttributeList)


# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/ns-processthreadsapi-process_information
class PROCESS_INFORMATION(ctypes.Structure):
    _fields_ = [
        ("hProcess", HANDLE),
        ("hThread", HANDLE),
        ("dwProcessId", DWORD),
        ("dwThreadId", DWORD),
    ]


# https://learn.microsoft.com/en-us/windows/win32/api/profinfo/ns-profinfo-profileinfoa
class PROFILEINFO(ctypes.Structure):
    _fields_ = [
        ("dwSize", DWORD),
        ("dwFlags", DWORD),
        ("lpUserName", LPWSTR),
        ("lpProfilePath", LPWSTR),
        ("lpDefaultPath", LPWSTR),
        ("lpServerName", LPWSTR),
        ("lpPolicyPath", LPWSTR),
        ("hProfile", HANDLE),
    ]


# https://learn.microsoft.com/en-us/windows/win32/api/wtypesbase/ns-wtypesbase-security_attributes
class SECURITY_ATTRIBUTES(ctypes.Structure):
    _fields_ = [("nLength", DWORD), ("lpSecurityDescriptor", LPVOID), ("bInheritHandle", BOOL)]


# https://learn.microsoft.com/en-us/windows/win32/api/ntdef/ns-ntdef-luid
class LUID(ctypes.Structure):
    _fields_ = [("LowPart", ULONG), ("HighPart", LONG)]


# https://learn.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-luid_and_attributes
class LUID_AND_ATTRIBUTES(ctypes.Structure):
    _fields_ = [("Luid", LUID), ("Attributes", DWORD)]


# https://learn.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-token_privileges
class TOKEN_PRIVILEGES(ctypes.Structure):
    _fields_ = [
        ("PrivilegeCount", DWORD),
        # Note: To use
        #   ctypes.cast(ctypes.byref(self.Privileges), ctypes.POINTER(LUID_AND_ATTRIBUTES * self.PrivilegeCount)).contents
        ("Privileges", LUID_AND_ATTRIBUTES * 0),
    ]

    @staticmethod
    def allocate(length: int) -> "TOKEN_PRIVILEGES":
        malloc_size_in_bytes = ctypes.sizeof(TOKEN_PRIVILEGES) + length * ctypes.sizeof(
            LUID_AND_ATTRIBUTES
        )
        malloc_buffer = (ctypes.c_byte * malloc_size_in_bytes)()
        token_privs = ctypes.cast(malloc_buffer, POINTER(TOKEN_PRIVILEGES))[0]
        token_privs.PrivilegeCount = length
        return token_privs

    def privileges_array(self) -> Sequence[LUID_AND_ATTRIBUTES]:
        return ctypes.cast(
            ctypes.byref(self.Privileges), ctypes.POINTER(LUID_AND_ATTRIBUTES * self.PrivilegeCount)
        ).contents


# =======================
# APIs
# =======================

# ---------
# From: kernel32.dll
# ---------
kernel32 = ctypes.WinDLL("kernel32")

# https://learn.microsoft.com/en-us/windows/win32/api/handleapi/nf-handleapi-closehandle
kernel32.CloseHandle.restype = BOOL
kernel32.CloseHandle.argtypes = [HANDLE]  # [in] hObject

# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-deleteprocthreadattributelist
kernel32.DeleteProcThreadAttributeList.restype = None
kernel32.DeleteProcThreadAttributeList.argtypes = [
    LPPROC_THREAD_ATTRIBUTE_LIST,  # [in, out] lpAttributeList
]

# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-getcurrentprocess
kernel32.GetCurrentProcess.restype = HANDLE
kernel32.GetCurrentProcess.argtypes = []

# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-getcurrentprocessid
kernel32.GetCurrentProcessId.restype = DWORD
kernel32.GetCurrentProcessId.argtypes = []

# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-initializeprocthreadattributelist
kernel32.InitializeProcThreadAttributeList.restype = BOOL
kernel32.InitializeProcThreadAttributeList.argtypes = [
    LPPROC_THREAD_ATTRIBUTE_LIST,  # [out, optional] lpAttributeList
    DWORD,  # [in] dwAttributeCount
    DWORD,  # [in] dwFlags
    PSIZE_T,  # [in,out] lpSize
]

# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-updateprocthreadattribute
kernel32.UpdateProcThreadAttribute.restype = BOOL
kernel32.UpdateProcThreadAttribute.argtypes = [
    LPPROC_THREAD_ATTRIBUTE_LIST,  # [in,out] lpAttributeList
    DWORD,  # [in] dwFlags
    SIZE_T,  # [in] Attribute (note: Pointer-sized integer; not an actual pointer)
    c_void_p,  # [in] lpValue
    SIZE_T,  # [in] cbSize
    c_void_p,  # [out, optional] lpPreviousValue
    PSIZE_T,  # [in, optional] lpReturnSize
]

# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-processidtosessionid
kernel32.ProcessIdToSessionId.restype = BOOL
kernel32.ProcessIdToSessionId.argtypes = [
    DWORD,  # [in] dwProcessId
    PDWORD,  # [out] pSessionId
]

# exports:
CloseHandle = kernel32.CloseHandle
DeleteProcThreadAttributeList = kernel32.DeleteProcThreadAttributeList
GetCurrentProcess = kernel32.GetCurrentProcess
GetCurrentProcessId = kernel32.GetCurrentProcessId
ProcessIdToSessionId = kernel32.ProcessIdToSessionId
InitializeProcThreadAttributeList = kernel32.InitializeProcThreadAttributeList
UpdateProcThreadAttribute = kernel32.UpdateProcThreadAttribute

# ---------
# From: advapi32.dll
# ---------
advapi32 = ctypes.WinDLL("advapi32")

# https://learn.microsoft.com/en-us/windows/win32/api/securitybaseapi/nf-securitybaseapi-adjusttokenprivileges
advapi32.AdjustTokenPrivileges.restype = BOOL
advapi32.AdjustTokenPrivileges.argtypes = [
    HANDLE,  # [in] TokenHandle
    BOOL,  # [in] DisableAllPrivileges
    POINTER(TOKEN_PRIVILEGES),  # [in, optional] NewState
    DWORD,  # [in] BufferLength
    POINTER(TOKEN_PRIVILEGES),  # [out, optional] PreviousState
    PDWORD,  # [out, optional] ReturnLength
]

# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-createprocessasuserw
advapi32.CreateProcessAsUserW.restype = BOOL
advapi32.CreateProcessAsUserW.argtypes = [
    HANDLE,  # [in, optional] hToken
    LPCWSTR,  # [in, optional] lpApplicationName
    LPWSTR,  # [in, out, optional] lpCommandLine
    POINTER(SECURITY_ATTRIBUTES),  # [in, optional] lpProcessAttributes
    POINTER(SECURITY_ATTRIBUTES),  # [in, optional] lpThreadAttributes
    BOOL,  # [in] bInheritHandles
    DWORD,  # [in] dwCreationFlags
    LPVOID,  # [in, optional] lpEnvironment
    LPCWSTR,  # [in, optional] lpCurrentDirectory
    POINTER(STARTUPINFOEX),  # [in] lpStartupInfo
    POINTER(PROCESS_INFORMATION),  # [out] lpProcessInformation
]

# https://learn.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-createprocesswithlogonw
advapi32.CreateProcessWithLogonW.restype = BOOL
advapi32.CreateProcessWithLogonW.argtypes = [
    LPCWSTR,  # [in] lpUsername
    LPCWSTR,  # [in, optional] lpDomain
    LPCWSTR,  # [in] lpPassword
    DWORD,  # [in] dwLogonFlags
    LPCWSTR,  # [in, optional] lpApplicationName
    LPWSTR,  # [in, out, optional] lpCommandLine
    DWORD,  # [in] dwCreationFlags
    LPVOID,  # [in, optional] lpEnvironment
    LPCWSTR,  # [in, optional] lpCurrentDirectory
    POINTER(STARTUPINFO),  # [in] lpStartupInfo
    POINTER(PROCESS_INFORMATION),  # [out] lpProcessInformation
]

# https://learn.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-createprocesswithtokenw
advapi32.CreateProcessWithTokenW.restype = BOOL
advapi32.CreateProcessWithTokenW.argtypes = [
    HANDLE,  # [in] hToken
    DWORD,  # [in] dwLogonFlags
    LPCWSTR,  # [in, optional] lpApplicationName
    LPWSTR,  # [in, out, optional] lpCommandLine
    DWORD,  # [in] dwCreationFlags
    LPVOID,  # [in,  optional] lpEnvironment
    LPCWSTR,  # [in, optional]  lpCurrentDirectory
    POINTER(STARTUPINFO),  # [in] lpStartupInfo
    POINTER(PROCESS_INFORMATION),  # [out] lpProcessInformation
]


# https://learn.microsoft.com/en-us/windows/win32/api/securitybaseapi/nf-securitybaseapi-gettokeninformation
advapi32.GetTokenInformation.restype = BOOL
advapi32.GetTokenInformation.argtypes = [
    HANDLE,  # [in] TokenHandle
    DWORD,  # [in] TokenInformationClass (actually an enum)
    LPVOID,  # [out, optional] TokenInformation
    DWORD,  # [in] TokenInformationLength
    PDWORD,  # [out] ReturnLength
]

# https://learn.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-logonuserw
advapi32.LogonUserW.restype = BOOL
advapi32.LogonUserW.argtypes = [
    LPCWSTR,  # [in] lpszUsername
    LPCWSTR,  # [in, optional] lpszDomain
    LPCWSTR,  # [in, optional] lpszPassword
    DWORD,  # [in] dwLogonType
    DWORD,  # [in] dwLogonProvider
    PHANDLE,  # [out] phToken
]

# https://learn.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-lookupprivilegenamew
advapi32.LookupPrivilegeNameW.restype = BOOL
advapi32.LookupPrivilegeNameW.argtypes = [
    LPCWSTR,  # [in, optional] lpSystemName
    POINTER(LUID),  # [in] lpLuid
    LPWSTR,  # [out] lpName
    LPDWORD,  # [in, out] cchName
]

# https://learn.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-lookupprivilegevaluew
advapi32.LookupPrivilegeValueW.restype = BOOL
advapi32.LookupPrivilegeValueW.argtypes = [
    LPCWSTR,  # [in, optional] lpSystemName
    LPCWSTR,  # [in] lpName
    POINTER(LUID),  # [out] lpLuid
]

# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-openprocesstoken
advapi32.OpenProcessToken.restype = BOOL
advapi32.OpenProcessToken.argtypes = [
    HANDLE,  # [in] ProcessHandle,
    DWORD,  # [in] DesiredAccess
    PHANDLE,  # [out] TokenHandle
]


# exports:
AdjustTokenPrivileges = advapi32.AdjustTokenPrivileges
CreateProcessAsUserW = advapi32.CreateProcessAsUserW
CreateProcessWithLogonW = advapi32.CreateProcessWithLogonW
CreateProcessWithTokenW = advapi32.CreateProcessWithTokenW
GetTokenInformation = advapi32.GetTokenInformation
LogonUserW = advapi32.LogonUserW
LookupPrivilegeNameW = advapi32.LookupPrivilegeNameW
LookupPrivilegeValueW = advapi32.LookupPrivilegeValueW
OpenProcessToken = advapi32.OpenProcessToken

# ---------
# From: userenv.dll
# ---------
userenv = ctypes.WinDLL("userenv")

# https://learn.microsoft.com/en-us/windows/win32/api/userenv/nf-userenv-createenvironmentblock
userenv.CreateEnvironmentBlock.restype = BOOL
userenv.CreateEnvironmentBlock.argtypes = [
    POINTER(ctypes.c_void_p),  # [in] lpEnvironment
    HANDLE,  # [in, optional] hToken
    BOOL,  # [in] bInherit
]

# # https://learn.microsoft.com/en-us/windows/win32/api/userenv/nf-userenv-destroyenvironmentblock
userenv.DestroyEnvironmentBlock.restype = BOOL
userenv.DestroyEnvironmentBlock.argtypes = [
    ctypes.c_void_p,  # [in] lpEnvironment
]

# https://learn.microsoft.com/en-us/windows/win32/api/userenv/nf-userenv-expandenvironmentstringsforuserw
userenv.ExpandEnvironmentStringsForUserW.restype = BOOL
userenv.ExpandEnvironmentStringsForUserW.argtypes = [
    HANDLE,  # [in, optional] hToken
    LPCWSTR,  # [in] lpSrc
    LPWSTR,  # [out] lpDest
    DWORD,  # [in] dwSize
]

# https://learn.microsoft.com/en-us/windows/win32/api/userenv/nf-userenv-loaduserprofilew
userenv.LoadUserProfileW.restype = BOOL
userenv.LoadUserProfileW.argtypes = [
    HANDLE,  # [in] hToken
    POINTER(PROFILEINFO),  # [in, out] lpProfileInfo
]

# https://learn.microsoft.com/en-us/windows/win32/api/userenv/nf-userenv-unloaduserprofile
userenv.UnloadUserProfile.restype = BOOL
userenv.UnloadUserProfile.argtypes = [
    HANDLE,  # [in] hToken
    HANDLE,  # [in] hProfile
]

# exports:
CreateEnvironmentBlock = userenv.CreateEnvironmentBlock
DestroyEnvironmentBlock = userenv.DestroyEnvironmentBlock
ExpandEnvironmentStringsForUserW = userenv.ExpandEnvironmentStringsForUserW
LoadUserProfileW = userenv.LoadUserProfileW
UnloadUserProfile = userenv.UnloadUserProfile



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import shlex
import signal
import sys
import time
from contextlib import nullcontext
from datetime import timedelta
from pathlib import Path
from queue import Queue, Empty
from subprocess import DEVNULL, PIPE, STDOUT, Popen, list2cmdline, run
from threading import Event, Thread
from typing import Any
from typing import Callable, Literal, Optional, Sequence, cast

from ._linux._capabilities import try_use_cap_kill
from ._linux._sudo import find_sudo_child_process_group_id
from ._logging import LoggerAdapter, LogContent, LogExtraInfo
from ._os_checker import is_linux, is_posix, is_windows
from ._session_user import PosixSessionUser, WindowsSessionUser, SessionUser

if is_windows():  # pragma: nocover
    from subprocess import CREATE_NEW_PROCESS_GROUP, CREATE_NO_WINDOW  # type: ignore
    from ._win32._popen_as_user import PopenWindowsAsUser  # type: ignore
    from ._windows_process_killer import kill_windows_process_tree


__all__ = ("LoggingSubprocess",)

# ========================================================================
# ========================================================================
# DEVELOPER NOTE:
#  If you make changes to this class's implementation, then be sure to test
# the cross-user functionality to make sure that is intact. The
# scripts/run_sudo_tests.sh script in this repository can be used to
# run these tests on Linux.
# ========================================================================
# ========================================================================

POSIX_SIGNAL_SUBPROC_SCRIPT_PATH = (
    Path(__file__).parent / "_scripts" / "_posix" / "_signal_subprocess.sh"
)

WINDOWS_SIGNAL_SUBPROC_SCRIPT_PATH = (
    Path(__file__).parent / "_scripts" / "_windows" / "_signal_win_subprocess.py"
)

LOG_LINE_MAX_LENGTH = 64 * 1000  # Start out with 64 KB, can increase if needed
STDOUT_END_GRACETIME_SECONDS = 5


class LoggingSubprocess(object):
    """A process whose stdout/stderr lines are sent to a given Logger."""

    _logger: LoggerAdapter
    _process: Optional[Popen]
    _args: Sequence[str]
    _encoding: str
    _user: Optional[SessionUser]
    _callback: Optional[Callable[[], None]]
    _start_failed: bool
    _has_started: Event
    _os_env_vars: Optional[dict[str, Optional[str]]]
    _working_dir: Optional[str]

    _pid: Optional[int]
    _sudo_child_process_group_id: Optional[int]
    _returncode: Optional[int]

    def __init__(
        self,
        *,
        logger: LoggerAdapter,
        args: Sequence[str],
        encoding: str = "utf-8",
        user: Optional[SessionUser] = None,  # OS-user to run as
        callback: Optional[Callable[[], None]] = None,
        os_env_vars: Optional[dict[str, Optional[str]]] = None,
        working_dir: Optional[str] = None,
    ):
        if len(args) < 1:
            raise ValueError("'args' kwarg must be a sequence of at least one element")
        if user is not None and os.name == "posix" and not isinstance(user, PosixSessionUser):
            raise ValueError("Argument 'user' must be a PosixSessionUser on posix systems.")
        if user is not None and is_windows() and not isinstance(user, WindowsSessionUser):
            raise ValueError("Argument 'user' must be a WindowsSessionUser on Windows systems.")

        self._logger = logger
        self._args = args[:]  # Make a copy
        self._encoding = encoding
        self._user = user
        self._callback = callback
        self._process = None
        self._os_env_vars = os_env_vars
        self._working_dir = working_dir
        self._start_failed = False
        self._has_started = Event()
        self._pid = None
        self._returncode = None
        self._sudo_child_process_group_id = None

    @property
    def pid(self) -> Optional[int]:
        return self._pid

    @property
    def exit_code(self) -> Optional[int]:
        """
        :return: None if the process has not yet exited. Otherwise, it returns the exit code of the
            process
        """
        # The process.wait() in the run() method ensures that the returncode
        # has been set once the subprocess has completed running. Don't poll here...
        # we only want to make the returncode available after the run method has
        # completed its work.
        if self._process is not None:
            return self._process.returncode
        return self._returncode

    @property
    def is_running(self) -> bool:
        """
        Determine whether the subprocess is running.
        :return: True if it is running; False otherwise
        """
        # Note: _process is None when either:
        #  a) The process failed to start; or
        #  b) The process has completed, and we've deleted the Popen instance
        return self._has_started.is_set() and self._process is not None

    @property
    def has_started(self) -> bool:
        """Determine whether or not the subprocess has been started yet or not"""
        return self._has_started.is_set()

    @property
    def failed_to_start(self) -> bool:
        """Determine whether the subprocess failed to start."""
        return self._start_failed

    def wait_until_started(self, timeout: Optional[timedelta] = None) -> None:
        """Blocks the caller until the subprocess has been started
        and is either running or has failed to start running.
        Args:
           timeout - Cease waiting after the given number of seconds has elapsed.
        """
        self._has_started.wait(timeout.total_seconds() if timeout is not None else None)

    def run(self) -> None:
        """Run the subprocess. The subprocess cannot be run if it has already been run, or is
        running.
        This is a blocking call.
        """
        if self._has_started.is_set():
            raise RuntimeError("The process has already been run")

        self._process = self._start_subprocess()
        # Set _has_started regardless of whether we started the process successfully or
        # not. That will wake up anyone waiting on wait_until_started() to know whether
        # we've gotten this far.
        self._has_started.set()
        if self._process is None:
            # We failed to start the subprocess
            self._start_failed = True
            if self._callback:
                self._callback()
            return

        self._pid = self._process.pid

        # Would use is_posix(), but it doesn't short-circuit mypy which then complains
        # about os.getpgid not being a valid attribute.
        if not sys.platform == "win32":
            if not self._user or self._user.is_process_user():
                self._sudo_child_process_group_id = os.getpgid(self._process.pid)
            else:
                self._sudo_child_process_group_id = find_sudo_child_process_group_id(
                    logger=self._logger,
                    sudo_process=self._process,
                )

        self._logger.info(
            f"Command started as pid: {self._process.pid}",
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )
        self._logger.info(
            "Output:",
            extra=LogExtraInfo(openjd_log_content=LogContent.BANNER | LogContent.COMMAND_OUTPUT),
        )

        try:
            self._log_subproc_stdout()  # Blocking
            self._process.wait()
            self._returncode = self._process.returncode
            self._log_returncode()
            if self._callback:
                self._callback()
        finally:
            # Explicitly delete the Popen in case it's a PopenWindowsAsUser and there's stuff to
            # deallocate
            proc = self._process
            self._process = None
            del proc

    def notify(self) -> None:
        """The 'Notify' part of Open Job Description's subprocess cancelation method.
        On Linux/macOS:
            - Send a SIGTERM to the parent process
        On Windows:
            - Send a CTRL_BREAK_EVENT to the process group

        TODO: Send the signal to every direct and transitive child of the parent
        process.
        """
        if self._process is not None and self._process.poll() is None:
            if is_posix():
                self._posix_signal_subprocess(signal_name="term")
            else:
                self._windows_notify_subprocess()

    def terminate(self) -> None:
        """The 'Terminate' part of Open Job Description's subprocess cancelation method.
        On Linux/macOS:
            - Send a SIGKILL to the parent process
        On Windows:
            - Not yet supported.

        TODO: Send the signal to every direct and transitive child of the parent
        process.
        """
        if self._process is not None and self._process.poll() is None:
            if is_posix():
                self._posix_signal_subprocess(signal_name="kill")
            else:
                self._logger.info(
                    f"INTERRUPT: Start killing the process tree with the root pid: {self._process.pid}",
                    extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
                )
                kill_windows_process_tree(self._logger, self._process.pid, signal_subprocesses=True)

    def _start_subprocess(self) -> Optional[Popen]:
        """Helper invoked by self.run() to start up the subprocess."""
        try:
            command: list[str] = []
            if self._user is not None:
                if is_posix():
                    user = cast(PosixSessionUser, self._user)
                    # Only sudo if the user to run as is not the same as the current user.
                    if not user.is_process_user():
                        # Note: setsid is required; else the running process will be in the
                        # same process group as the `sudo` command. If that happens, then
                        # we're stuck: 1/ Our user cannot kill processes by the self._user; and
                        # 2/ The self._user cannot kill the root-owned sudo process group.
                        command.extend(["sudo", "-u", user.user, "-i", "setsid", "-w"])
                elif is_windows():
                    user = cast(WindowsSessionUser, self._user)  # type: ignore

            command.extend(self._args)

            # Append the given environment to the current one.
            popen_args: dict[str, Any] = dict(
                args=command,
                stdin=DEVNULL,
                stdout=PIPE,
                stderr=STDOUT,
                encoding=self._encoding,
                start_new_session=True,
                cwd=self._working_dir,
            )

            if is_windows():
                # We need a process group in order to send notify signals
                # https://docs.python.org/2/library/subprocess.html#subprocess.CREATE_NEW_PROCESS_GROUP
                popen_args["creationflags"] = CREATE_NEW_PROCESS_GROUP

            cmd_line_for_logger: str
            if is_posix():
                cmd_line_for_logger = shlex.join(command)
            else:
                cmd_line_for_logger = list2cmdline(self._args)
            self._logger.info(
                "Running command %s",
                cmd_line_for_logger,
                extra=LogExtraInfo(
                    openjd_log_content=LogContent.FILE_PATH | LogContent.PROCESS_CONTROL
                ),
            )

            process: Popen
            if is_windows() and self._user and not user.is_process_user():
                popen_args["env"] = self._os_env_vars
                process = PopenWindowsAsUser(user, **popen_args)  # type: ignore
            else:
                if self._os_env_vars:
                    # Our env vars may have 'None' as the value for some keys.
                    # Semantically, that means that we want to delete that environment
                    # variable's value from the environment. However, Popen doesn't do that;
                    # it will instead choose to blow-up if any env var value is None
                    env: dict[str, Optional[str]] = dict(os.environ)
                    env.update(**self._os_env_vars)
                    popen_env: dict[str, str] = {k: v for k, v in env.items() if v is not None}
                    popen_args["env"] = popen_env
                process = Popen(**popen_args)
            return process

        except Exception as e:
            self._logger.info(
                f"Process failed to start: {str(e)}",
                extra=LogExtraInfo(
                    openjd_log_content=LogContent.EXCEPTION_INFO | LogContent.PROCESS_CONTROL
                ),
            )
            return None

    def _log_subproc_stdout(self) -> None:
        """
        Blocking call which logs the STDOUT of the running subproc until the subprocess exits.

        Note that this can result in hanging threads if:
            1. The command we run creates a detatched grandchild process
            2. The detached grandchild inherits the STDOUT stream of the process we run
            3. The detached grandchild process does not write to STDOUT, ever.

        In the above situation, there is a thread leak due to there not being a unilaterally
        available python API which performs a timed-out blocking read on IO streams. If the
        grandchild process ever:
            1. Exits
            2. Writes to STDOUT
        or
            3. The python session running this code ends

        Then the thread exits and is cleaned up automatically by the python garbage collector.
        """
        assert self._process
        stream = self._process.stdout
        # Convince type checker that stdout is not None
        assert stream is not None

        exit_event = Event()

        # Process stdout/stderr of the job; echoing it to our logger
        def _stream_readline_max_length():
            if exit_event.is_set():
                return ""  # we can return anything here, just forces the iter to loop
            # Enforce a max line length for readline to ensure we don't infinitely grow the buffer
            return stream.readline(LOG_LINE_MAX_LENGTH)  # type: ignore

        stdout_queue: Queue[str] = Queue()

        def _enqueue_stdout():
            """
            Enqueues all the stdout from stream into the given queue, until the stream is closed
            or exit_event is set.
            """
            for line in iter(_stream_readline_max_length, ""):
                if exit_event.is_set():
                    break
                line = line.rstrip("\n\r")
                stdout_queue.put(line)
            stream.close()

        process_exit_time = None
        warn_time = None

        logging_thread = Thread(target=_enqueue_stdout, daemon=True)
        # We start the thread as a daemon, and explicitly do not call join() on it ever because:
        #      If the subprocess creates a child subprocess that inherits the STDOUT stream, then exits while leaving
        #      the child process running. There is an edge case where the child process never write to STDOUT and we are
        #      stuck waiting on stream.readline() to stop blocking before exit_event can be checked to end the thread.
        #      In this case we leave a dangling thread until the subprocess exits. This was done because
        #       1. Python does not have support for a blocking read with a timeout.
        #       2. Python does not have unilateral support for non-blocking reads until Python 3.12 and non-blocking
        #          reads would require arbitrary sleep calls while waiting for output which could cause a performance
        #          impact.
        logging_thread.start()

        while (
            logging_thread.is_alive()
        ):  # If the logging thread is alive, the STDOUT stream is still open
            try:
                # We timeout after 1 ms because the main process can end while leaving child processes that
                # prevent closing the STDOUT stream. Waiting a maximum of 1 ms allows us to detect this quickly, while
                # not significantly impacting CPU usage.
                line = stdout_queue.get(timeout=0.001)
                self._logger.info(
                    line, extra=LogExtraInfo(openjd_log_content=LogContent.COMMAND_OUTPUT)
                )
            except Empty:
                pass  # queue.get timed out. This means the subprocess does not print much to STDOUT. Just continue.

            if self._process.poll() is not None:  # The main command exited.
                if process_exit_time is None:
                    process_exit_time = time.monotonic()
                elif (time.monotonic() - process_exit_time) < 1:
                    # There could be a bunch of STDOUT buffered up. We don't want the warning to be too noisy, so
                    # give a second to clear the queue before we get stern and warn about ending the action.
                    continue
                elif not warn_time:
                    # It's been over a second of trying to empty STDOUT. Most likely the stream is still open.
                    self._logger.warning(
                        f"Command exited but STDOUT stream is still open. Waiting gracetime of {STDOUT_END_GRACETIME_SECONDS} seconds for the STDOUT stream to close before ending action.",
                        extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
                    )
                    warn_time = time.monotonic()
                elif (time.monotonic() - warn_time) > STDOUT_END_GRACETIME_SECONDS:
                    self._logger.warning(
                        f"Gracetime of {STDOUT_END_GRACETIME_SECONDS} seconds elapsed but the STDOUT stream is still open. Ending action.",
                        extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
                    )
                    exit_event.set()  # When the STDOUT stream ends this will cause the thread to exit.
                    break

        while not stdout_queue.empty():
            # empty the queue
            line = stdout_queue.get()
            self._logger.info(
                line, extra=LogExtraInfo(openjd_log_content=LogContent.COMMAND_OUTPUT)
            )

    def _log_returncode(self):
        """Logs the return code of the exited subprocess"""
        if self._returncode is not None:
            # Print out the signed representation of returncodes that would be negative as a 32-bit signed integer
            if self._returncode < 0x7FFFFFFF:
                self._logger.info(
                    f"Process pid {self._process.pid} exited with code: {self._returncode} (unsigned) / {hex(self._returncode)} (hex)",
                    extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
                )
            else:

                def _tosigned(n: int) -> int:
                    b = (n & 0xFFFFFFFF).to_bytes(4, "big", signed=False)
                    return int.from_bytes(b, "big", signed=True)

                self._logger.info(
                    f"Process pid {self._process.pid} exited with code: {self._returncode} (unsigned) / {hex(self._returncode)} (hex) / {_tosigned(self._returncode)} (signed)",
                    extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
                )

    def _posix_signal_subprocess(
        self,
        signal_name: Literal["term", "kill"],
    ) -> None:
        """Send a given named signal to the subprocess."""

        # Hint to mypy to not raise module attribute errors (e.g. missing os.getpgid)
        if sys.platform == "win32":
            raise NotImplementedError("This method is for POSIX hosts only")

        # We can run into a race condition where the process exits (and another thread sets self._process to None)
        # before the cancellation happens, so we swap to a local variable to ensure a cancellation that is not needed,
        # does not raise an exception here.
        process = self._process
        # Convince the type checker that accessing process is okay
        assert process is not None

        # Note: A limitation of this implementation is that it will only sigkill
        # processes that are in the same process-group as the command that we ran.
        # In the future, we can extend this to killing all processes spawned (including into
        # new process-groups since the parent-pid will allow the mapping)
        # by a depth-first traversal through the children. At each recursive
        # step we:
        #  1. SIGSTOP the process, so that it cannot create new subprocesses;
        #  2. Recurse into each child; and
        #  3. SIGKILL the process.
        # Things to watch for when doing so:
        #  a. PIDs can get reused; just because a pid was a child of a process at one point doesn't
        #     mean that it's still the same process when we recurse to it. So, check that the parent-pid
        #     of any child is still as expected before we signal it or collect its children.
        #  b. When we run the command using `sudo` then we need to either run code that does the whole
        #     algorithm as the other user, or `sudo` to send every process signal.

        numeric_signal = 0
        if signal_name == "term":
            numeric_signal = signal.SIGTERM
            # SIGTERM is the simpler case. In the cross-user sudo case, we can send a signal to the
            # sudo process and it will forward the signal. For the same-user case, the subprocess
            # is the one we want to signal.
            self._logger.info(
                f'INTERRUPT: Sending signal "{signal_name}" to process {process.pid}',
                extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
            )
            try:
                os.kill(process.pid, numeric_signal)
            except OSError:
                self._logger.warning(
                    f"INTERRUPT: Unable to send {signal_name} to {process.pid}",
                    extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
                )
            return
        elif signal_name == "kill":
            numeric_signal = signal.SIGKILL
        else:
            raise NotImplementedError(f"Unsupported signal: {signal_name}")

        kill_cmd = list[str]()

        if self._user is not None:
            user = cast(PosixSessionUser, self._user)
            # Only sudo if the user to run as is not the same as the current user.
            if not user.is_process_user():
                kill_cmd = ["sudo", "-u", user.user, "-i"]

        # If we were unable to detect sudo's child process PID after launching the
        # subprocess, we try again now
        if not self._sudo_child_process_group_id:
            self._sudo_child_process_group_id = find_sudo_child_process_group_id(
                logger=self._logger,
                sudo_process=process,
            )

        if not self._sudo_child_process_group_id:
            self._logger.warning(
                f"Failed to send signal '{signal_name}': Unable to determine child process of sudo",
                extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
            )
            return

        # Try directly signaling the process(es) first
        ctx_mgr = try_use_cap_kill() if is_linux() else nullcontext(enter_result=False)
        with ctx_mgr as has_cap_kill:
            if has_cap_kill or not self._user or self._user.is_process_user():
                try:
                    self._logger.info(
                        f'INTERRUPT: Sending signal "{signal_name}" to process group {self._sudo_child_process_group_id}',
                        extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
                    )
                    os.killpg(self._sudo_child_process_group_id, numeric_signal)
                except OSError:
                    self._logger.info(
                        "Could not directly send signal {signal_name} to {self._posix_signal_target.pid}, trying sudo.",
                        extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
                    )
                else:
                    return
            else:
                self._logger.info(
                    "Could not directly send signal {signal_name} to {process.pid}, trying sudo.",
                    extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
                )

        # Uncomment to visualize process tree when debugging tests
        # self._log_process_tree()

        kill_cmd.extend(
            [
                "kill",
                "-s",
                signal_name,
                "--",
                f"-{self._sudo_child_process_group_id}",
            ]
        )
        self._logger.info(
            f"INTERRUPT: Running: {shlex.join(kill_cmd)}",
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )
        result = run(
            kill_cmd,
            stdout=PIPE,
            stderr=STDOUT,
            stdin=DEVNULL,
        )
        if result.returncode != 0:
            self._logger.warning(
                f"Failed to send signal '{signal_name}' to PGID {self._sudo_child_process_group_id}: %s",
                result.stdout.decode("utf-8"),
                extra=LogExtraInfo(
                    openjd_log_content=LogContent.PROCESS_CONTROL | LogContent.EXCEPTION_INFO
                ),
            )

    def _log_process_tree(self) -> None:
        """A developer method to visualize the process tree including PIDs and PGIDs when debuging tests"""
        pstree_result = run(["pstree", "-pg"], stdout=PIPE, stderr=STDOUT, stdin=DEVNULL, text=True)
        self._logger.debug(
            f"pstree -pg output: {pstree_result.stdout}",
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )
        ps_result = run(["ps", "-ejH"], stdout=PIPE, stderr=STDOUT, stdin=DEVNULL, text=True)
        self._logger.debug(
            f"ps -ejH output:\n{ps_result.stdout}",
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )

    def _windows_notify_subprocess(self) -> None:
        """Sends a CTRL_BREAK_EVENT signal to the subprocess"""
        # Convince the type checker that accessing _process is okay
        assert self._process is not None

        # CTRL-C handler is disabled by default when CREATE_NEW_PROCESS_GROUP is passed.
        # We send CTRL-BREAK as handler for it cannnot be disabled.
        # https://learn.microsoft.com/en-us/windows/console/ctrl-c-and-ctrl-break-signals
        # https://learn.microsoft.com/en-us/windows/console/generateconsolectrlevent
        # https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-createprocessa#remarks
        # https://stackoverflow.com/questions/35772001/how-to-handle-a-signal-sigint-on-a-windows-os-machine/35792192#35792192
        self._logger.info(
            f"INTERRUPT: Sending CTRL_BREAK_EVENT to {self._process.pid}",
            extra=LogExtraInfo(openjd_log_content=LogContent.PROCESS_CONTROL),
        )

        # _process will be running in new console, we run another process to attach to it and send signal
        cmd = [
            # When running in a service context, we want to call the non-service Python binary
            sys.executable.lower().replace("pythonservice.exe", "python.exe"),
            str(WINDOWS_SIGNAL_SUBPROC_SCRIPT_PATH),
            str(self._process.pid),
        ]
        result = run(
            cmd,
            stdout=PIPE,
            stderr=STDOUT,
            stdin=DEVNULL,
            creationflags=CREATE_NEW_PROCESS_GROUP | CREATE_NO_WINDOW,
        )
        if result.returncode != 0:
            self._logger.warning(
                f"Failed to send signal 'CTRL_BREAK_EVENT' to subprocess {self._process.pid}: %s",
                result.stdout.decode("utf-8"),
                extra=LogExtraInfo(
                    openjd_log_content=LogContent.PROCESS_CONTROL | LogContent.EXCEPTION_INFO
                ),
            )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import sys

LINUX = "linux"
MACOS = "darwin"
POSIX = "posix"
WINDOWS = "nt"


def is_linux() -> bool:
    return sys.platform == LINUX


def is_posix() -> bool:
    return os.name == POSIX


def is_windows() -> bool:
    return os.name == WINDOWS


def check_os() -> None:
    if not (is_posix() or is_windows()):
        raise NotImplementedError(
            "Open Job Description can be only run in Posix or Windows system."
            f"os: {os.name} is not supported yet."
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from dataclasses import dataclass, fields
from enum import Enum
from os import name as os_name
from pathlib import PurePath, PurePosixPath, PureWindowsPath


class PathFormat(str, Enum):
    POSIX = "POSIX"
    WINDOWS = "WINDOWS"


@dataclass(frozen=True)
class PathMappingRule:
    source_path_format: PathFormat
    source_path: PurePath
    destination_path: PurePath

    def __init__(
        self, *, source_path_format: PathFormat, source_path: PurePath, destination_path: PurePath
    ):
        if source_path_format == PathFormat.POSIX:
            if not isinstance(source_path, PurePosixPath):
                raise ValueError(
                    "Path mapping rule source_path_format does not match source_path type"
                )
        else:
            if not isinstance(source_path, PureWindowsPath):
                raise ValueError(
                    "Path mapping rule source_path_format does not match source_path type"
                )

        # This roundabout way can set the attributes of a frozen dataclass
        object.__setattr__(self, "source_path_format", source_path_format)
        object.__setattr__(self, "source_path", source_path)
        object.__setattr__(self, "destination_path", destination_path)

    @staticmethod
    def from_dict(rule: dict[str, str]) -> "PathMappingRule":
        """Builds a PathMappingRule from a dictionary representation
        with strings as values."""
        if not rule:
            raise ValueError("Empty path mapping rule")

        field_names = [field.name for field in fields(PathMappingRule)]
        for name in field_names:
            if name not in rule:
                raise ValueError(f"Path mapping rule requires the following fields: {field_names}")

        source_path_format = PathFormat(rule["source_path_format"].upper())
        source_path: PurePath
        if source_path_format == PathFormat.POSIX:
            source_path = PurePosixPath(rule["source_path"])
        else:
            source_path = PureWindowsPath(rule["source_path"])
        destination_path = PurePath(rule["destination_path"])

        unsupported_fields = set(rule.keys()) - set(field_names)
        if unsupported_fields:
            raise ValueError(
                f"Unsupported fields for constructing path mapping rule: {unsupported_fields}"
            )

        return PathMappingRule(
            source_path_format=source_path_format,
            source_path=source_path,
            destination_path=destination_path,
        )

    def to_dict(self) -> dict[str, str]:
        """Returns a dictionary representation of the PathMappingRule."""
        return {
            "source_path_format": self.source_path_format.name,
            "source_path": str(self.source_path),
            "destination_path": str(self.destination_path),
        }

    def apply(self, *, path: str) -> tuple[bool, str]:
        """Applies the path mapping rule on the given path, if it matches the rule.
        Does not collapse ".." since symbolic paths could be used.

        Returns: tuple[bool, str] - indicating if the path matched the rule and the resulting
        mapped path. If it doesn't match, then it returns the original path unmodified.
        """
        pure_path: PurePath
        if self.source_path_format == PathFormat.POSIX:
            pure_path = PurePosixPath(path)
        else:
            pure_path = PureWindowsPath(path)

        if not pure_path.is_relative_to(self.source_path):
            return False, path

        remapped_parts = (
            self.destination_path.parts + pure_path.parts[len(self.source_path.parts) :]
        )
        if os_name == "posix":
            result = str(PurePosixPath(*remapped_parts))
            if self._has_trailing_slash(self.source_path_format, path):
                result += "/"
        else:
            result = str(PureWindowsPath(*remapped_parts))
            if self._has_trailing_slash(self.source_path_format, path):
                result += "\\"

        return True, result

    def _has_trailing_slash(self, os: PathFormat, path: str) -> bool:
        if os == PathFormat.POSIX:
            return path.endswith("/")
        else:
            return path.endswith("\\")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import sys

# This assertion short-circuits mypy from type checking this module on platforms other than Windows
# https://mypy.readthedocs.io/en/stable/common_issues.html#python-version-and-system-platform-checks
assert sys.platform == "win32"

from typing import Any, Optional, cast
import ctypes
from ctypes.wintypes import HANDLE
from subprocess import list2cmdline, Popen
from subprocess import Handle  # type: ignore # linter doesn't know it exists
import platform
from ._api import (
    # Constants
    LOGON_WITH_PROFILE,
    PROC_THREAD_ATTRIBUTE_HANDLE_LIST,
    STARTF_USESHOWWINDOW,
    STARTF_USESTDHANDLES,
    SW_HIDE,
    # Structures
    PROCESS_INFORMATION,
    STARTUPINFO,
    STARTUPINFOEX,
    # Functions
    CloseHandle,
    CreateProcessAsUserW,
    CreateProcessWithLogonW,
    DestroyEnvironmentBlock,
    UpdateProcThreadAttribute,
)
from ._helpers import (
    environment_block_for_user,
    # environment_block_for_user_context,
    environment_block_from_dict,
    environment_block_to_dict,
    logon_user_context,
)
from .._session_user import WindowsSessionUser

if platform.python_implementation() != "CPython":
    raise RuntimeError(
        f"Not compatible with the {platform.python_implementation} of Python. Please use CPython."
    )

CREATE_UNICODE_ENVIRONMENT = 0x00000400
EXTENDED_STARTUPINFO_PRESENT = 0x00080000


def inherit_handles(startup_info: STARTUPINFOEX, handles: tuple[int]) -> ctypes.Array:
    """Set the given 'startup_info' to have the subprocess inherit the given handles, and only the
    given handles.

    Returns:
        - The allocated list of handles. Per the Win32 APIs, you must ensure that this buffer is
          only deallocated *after* the attribute list in the startup_info has been deallocated.
    """
    handles_list = (HANDLE * len(handles))()
    for i, h in enumerate(handles):
        handles_list[i] = h
    if not UpdateProcThreadAttribute(
        startup_info.lpAttributeList,
        0,  # reserved and must be 0
        PROC_THREAD_ATTRIBUTE_HANDLE_LIST,
        ctypes.byref(handles_list),
        ctypes.sizeof(handles_list),
        None,  # reserved and must be null
        None,  # reserved and must be null
    ):
        raise ctypes.WinError()
    return handles_list


class PopenWindowsAsUser(Popen):
    """Class to run a process as another user on Windows.
    Derived from Popen, it defines the _execute_child() method to call CreateProcessWithLogonW.
    """

    def __init__(self, user: WindowsSessionUser, *args: Any, **kwargs: Any):
        """
        Arguments:
            username (str):  Name of user to run subprocess as
            password (str):  Password for username
            args (Any):  Popen constructor args
            kwargs (Any):  Popen constructor kwargs
            https://docs.python.org/3/library/subprocess.html#popen-constructor
        """
        self.user = user
        super(PopenWindowsAsUser, self).__init__(*args, **kwargs)

    def _execute_child(
        self,
        args,
        executable,
        preexec_fn,
        close_fds,
        pass_fds,
        cwd,
        env,
        startupinfo,
        creationflags,
        shell,
        p2cread,
        p2cwrite,
        c2pread,
        c2pwrite,
        errread,
        errwrite,
        restore_signals,
        start_new_session,
        *additional_args,
        **kwargs,
    ):
        """Execute program (MS Windows version).
        Calls CreateProcessWithLogonW to run a process as another user.
        """

        assert not pass_fds, "pass_fds not supported on Windows."

        commandline = args if isinstance(args, str) else list2cmdline(args)
        # CreateProcess* may modify the commandline, so copy it to a mutable buffer
        cmdline = ctypes.create_unicode_buffer(commandline)

        if executable is not None:
            executable = os.fsdecode(executable)

        if cwd is not None:
            cwd = os.fsdecode(cwd)

        # Initialize structures
        si = STARTUPINFO()
        si.cb = ctypes.sizeof(STARTUPINFO)
        pi = PROCESS_INFORMATION()

        use_std_handles = -1 not in (p2cread, c2pwrite, errwrite)
        if use_std_handles:
            si.hStdInput = int(p2cread)
            si.hStdOutput = int(c2pwrite)
            si.hStdError = int(errwrite)
            si.dwFlags |= STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW
            # Ensure that the console window is hidden
            si.wShowWindow = SW_HIDE

        sys.audit("subprocess.Popen", executable, args, cwd, env, self.user.user)

        def _merge_environment(
            user_env: ctypes.c_void_p, env: dict[str, Optional[str]]
        ) -> ctypes.c_wchar_p:
            user_env_dict = cast(dict[str, Optional[str]], environment_block_to_dict(user_env))
            # All environment keys are converted to uppercase. This is because while Windows environment variables are case insensitive, the win32 api
            # allows you to store multiple keys of mixed case, for example the win32 api would allow you
            # to set both "PATH" and "Path" as environment variable keys. This leads to undefined behaviour when calling one of the keys.
            user_env_dict = {k.upper(): v for k, v in user_env_dict.items()}
            upper_cased_env = {k.upper(): v for k, v in env.items()}
            user_env_dict.update(**upper_cased_env)
            result = {k: v for k, v in user_env_dict.items() if v is not None}
            return environment_block_from_dict(result)

        env_ptr = ctypes.c_void_p(0)
        try:
            if self.user.password is not None:
                with logon_user_context(self.user.user, self.user.password) as logon_token:
                    env_ptr = environment_block_for_user(logon_token)
                    if env:
                        env_block = _merge_environment(env_ptr, env)
                    else:
                        env_block = env_ptr

                # https://learn.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-createprocesswithlogonw
                if not CreateProcessWithLogonW(
                    self.user.user,
                    None,  # TODO: Domains not yet supported
                    self.user.password,
                    LOGON_WITH_PROFILE,
                    executable,
                    cmdline,
                    creationflags | CREATE_UNICODE_ENVIRONMENT,
                    env_block,
                    cwd,
                    ctypes.byref(si),
                    ctypes.byref(pi),
                ):
                    # Raises: OSError
                    raise ctypes.WinError()
            elif self.user.logon_token is not None:
                siex = STARTUPINFOEX()
                ctypes.memmove(
                    ctypes.pointer(siex.StartupInfo), ctypes.pointer(si), ctypes.sizeof(STARTUPINFO)
                )
                siex.StartupInfo.cb = ctypes.sizeof(STARTUPINFOEX)
                creationflags |= EXTENDED_STARTUPINFO_PRESENT

                handles_list: Optional[ctypes.Array] = None
                handles_to_inherit = tuple(int(h) for h in (p2cread, c2pwrite, errwrite) if h != -1)
                try:
                    # Allocate the lpAttributeList array of the STARTUPINFOEX structure so that it
                    # has sufficient space for a single attribute. We only have a single attribute that
                    # we're setting -- namely a PROC_THREAD_ATTRIBUTE_HANDLE_LIST that itself contains a list of handles --
                    # so this is sufficient.
                    siex.allocate_attribute_list(1)

                    # Note: We must ensure that 'handles_list' must persist until the
                    # attribute list is destroyed using DeleteProcThreadAttributeList. We do this by holding on
                    # to a reference to it until after the finally block of this try.
                    handles_list = inherit_handles(  # noqa: F841 # ignore: assigned but not used
                        siex, handles_to_inherit
                    )

                    # From https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-createprocessasuserw
                    # If the lpEnvironment parameter is NULL, the new process inherits the environment of the calling process.
                    # CreateProcessAsUser does not automatically modify the environment block to include environment variables specific to
                    # the user represented by hToken. For example, the USERNAME and USERDOMAIN variables are inherited from the calling
                    # process if lpEnvironment is NULL. It is your responsibility to prepare the environment block for the new process and
                    # specify it in lpEnvironment.

                    env_ptr = environment_block_for_user(self.user.logon_token)
                    if env:
                        env_block = _merge_environment(env_ptr, env)
                    else:
                        env_block = env_ptr

                    if not CreateProcessAsUserW(
                        self.user.logon_token,
                        executable,
                        cmdline,
                        None,
                        None,
                        True,
                        creationflags | CREATE_UNICODE_ENVIRONMENT,
                        env_block,
                        cwd,
                        ctypes.byref(siex),
                        ctypes.byref(pi),
                    ):
                        # Raises: OSError
                        raise ctypes.WinError()
                finally:
                    siex.deallocate_attribute_list()
            else:
                raise NotImplementedError("Unexpected case for WindowsSessionUser properties")
        finally:
            if env_ptr.value is not None:
                DestroyEnvironmentBlock(env_ptr)
            # Child is launched. Close the parent's copy of those pipe
            # handles that only the child should have open.
            self._close_pipe_fds(p2cread, p2cwrite, c2pread, c2pwrite, errread, errwrite)

        # Retain the process handle, but close the thread handle
        CloseHandle(pi.hThread)

        self._child_created = True
        self.pid = pi.dwProcessId
        self._handle = Handle(pi.hProcess)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from ._os_checker import is_windows

if is_windows():
    import win32security
    import ntsecuritycon


class WindowsPermissionHelper:
    """
    This class contains helper methods to set permissions for files and directories on Windows.
    """

    @staticmethod
    def set_permissions(
        file_path: str,
        *,
        principals_full_control: list[str] = [],
        principals_modify_access: list[str] = [],
    ):
        """
        Grants access control over the object at file_path.
        Sets flags so both child files and directories inherit these permissions.

        Arguments:
            file_path (str): The path to the file or directory.
            principals_full_control (List[str]): The names of the principals to permit Full Control access.
            principals_modify_access (List[str]): The names of the principals to permit Modify access

        Raises:
            RuntimeError if there is a problem modifying the security attributes.
        """
        try:
            # We don't want to propagate existing permissions, so create a new DACL
            dacl = win32security.ACL()
            for principal in principals_full_control:
                user_or_group_sid, _, _ = win32security.LookupAccountName(None, principal)

                # Add an ACE to the DACL giving the principal full control and enabling inheritance of the ACE
                dacl.AddAccessAllowedAceEx(
                    win32security.ACL_REVISION,
                    ntsecuritycon.OBJECT_INHERIT_ACE | ntsecuritycon.CONTAINER_INHERIT_ACE,
                    ntsecuritycon.FILE_ALL_ACCESS,  # = 0x1F01FF
                    user_or_group_sid,
                )
            for principal in principals_modify_access:
                user_or_group_sid, _, _ = win32security.LookupAccountName(None, principal)

                # Add an ACE to the DACL giving the principal full control and enabling inheritance of the ACE
                dacl.AddAccessAllowedAceEx(
                    win32security.ACL_REVISION,
                    ntsecuritycon.OBJECT_INHERIT_ACE | ntsecuritycon.CONTAINER_INHERIT_ACE,
                    # Values of these constants defined in winnt.h
                    # Constant value after ORs: 0x1301FF
                    # Delta from FILE_ALL_ACCESS is 0xC0000 = WRITE_DAC(0x40000) | WRITE_OWNER(0x80000)
                    ntsecuritycon.FILE_GENERIC_READ  # = 0x120089
                    | ntsecuritycon.FILE_GENERIC_WRITE  # = 0x120116
                    | ntsecuritycon.FILE_GENERIC_EXECUTE  # = 0x1200A0
                    | ntsecuritycon.DELETE  # = 0x10000
                    | ntsecuritycon.FILE_DELETE_CHILD,  # = 0x0040
                    user_or_group_sid,
                )

            # Get the security descriptor of the tempdir
            sd = win32security.GetFileSecurity(
                str(file_path), win32security.DACL_SECURITY_INFORMATION
            )

            # Set the security descriptor's DACL to the newly-created DACL
            # Arguments:
            # 1. bDaclPresent = 1: Indicates that the DACL is present in the security descriptor.
            #    If set to 0, this method ignores the provided DACL and allows access to all principals.
            # 2. dacl: The discretionary access control list (DACL) to be set in the security descriptor.
            # 3. bDaclDefaulted = 0: Indicates the DACL was provided and not defaulted.
            #    If set to 1, indicates the DACL was defaulted, as in the case of permissions inherited from a parent directory.
            sd.SetSecurityDescriptorDacl(1, dacl, 0)

            # Set the security descriptor to the tempdir
            # Note: This completely overwrites the DACL; so, if we don't provide a permission above then
            # the DACL doesn't have it.
            win32security.SetFileSecurity(
                str(file_path), win32security.DACL_SECURITY_INFORMATION, sd
            )
        except Exception as err:
            raise RuntimeError(
                f"Could not change permissions of directory '{str(dir)}' (error: {str(err)})"
            )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import sys

# This assertion short-circuits mypy from type checking this module on platforms other than Windows
# https://mypy.readthedocs.io/en/stable/common_issues.html#python-version-and-system-platform-checks
assert sys.platform == "win32"

import win32api
import win32con
from ctypes.wintypes import DWORD, HANDLE
from ctypes import (
    WinError,
    byref,
    cast,
    c_void_p,
    c_wchar,
    c_wchar_p,
    sizeof,
)
from contextlib import contextmanager
from typing import Generator, Optional

from ._api import (
    # Constants
    LOGON32_LOGON_INTERACTIVE,
    LOGON32_PROVIDER_DEFAULT,
    # Functions
    CloseHandle,
    CreateEnvironmentBlock,
    DestroyEnvironmentBlock,
    GetCurrentProcessId,
    LogonUserW,
    ProcessIdToSessionId,
)


def get_process_user():
    """
    Returns the user name of the user running the current process.
    """
    return win32api.GetUserNameEx(win32con.NameSamCompatible)


def get_current_process_session_id() -> int:
    """
    Finds the Session ID of the current process, and returns it.
    """
    proc_id = GetCurrentProcessId()
    session_id = DWORD(0)
    # Ignore the return value; will only fail if given a bad
    # process id, and that's clearly impossible here.
    ProcessIdToSessionId(proc_id, byref(session_id))
    return session_id.value


def logon_user(username: str, password: str) -> HANDLE:
    """
    Attempt to logon as the given username & password.
    Return a HANDLE to a logon_token.

    Note:
      The caller *MUST* call CloseHandle on the returned value when done with it.
      Handles are not automatically garbage collected.

    Raises:
        OSError - If an error is encountered.
    """
    hToken = HANDLE(0)
    if not LogonUserW(
        username,
        None,  # TODO - domain handling??
        password,
        LOGON32_LOGON_INTERACTIVE,
        LOGON32_PROVIDER_DEFAULT,
        byref(hToken),
    ):
        raise WinError()

    return hToken


@contextmanager
def logon_user_context(username: str, password: str) -> Generator[HANDLE, None, None]:
    """
    A context manager wrapper around logon_user(). This will automatically
    Close the logon_token when the context manager is exited.
    """
    hToken: Optional[HANDLE] = None
    try:
        hToken = logon_user(username, password)
        yield hToken
    finally:
        if hToken is not None and not CloseHandle(hToken):
            raise WinError()


def environment_block_for_user(logon_token: HANDLE) -> c_void_p:
    """
    Create an Environment Block for a given logon_token and return it.
    Per https://learn.microsoft.com/en-us/windows/win32/api/userenv/nf-userenv-createenvironmentblock
    "The environment block is an array of null-terminated Unicode strings. The list ends with two nulls (\0\0)."

    Returns:
        Pointer to the environment block

    Raises:
        OSError - If there is an error creating the block

    Notes:
     1) The returned block *MUST* be deallocated with DestroyEnvironmentBlock when done
     2) Destroying an environment block while it is in use (e.g. while the process it was passed
        to is still running) WILL result in a hard to debug crash in ntdll.dll. So, don't do that!
    """
    environment = c_void_p()
    if not CreateEnvironmentBlock(byref(environment), logon_token, False):
        raise WinError()
    return environment


@contextmanager
def environment_block_for_user_context(logon_token: HANDLE) -> Generator[c_void_p, None, None]:
    """As environment_block_for_user, but as a context manager. This will automatically
    Destroy the environment block when exiting the context.
    """
    lp_environment: Optional[c_void_p] = None
    try:
        lp_environment = environment_block_for_user(logon_token)
        yield lp_environment
    finally:
        if lp_environment is not None and not DestroyEnvironmentBlock(lp_environment):
            raise WinError()


def environment_block_to_dict(block: c_void_p) -> dict[str, str]:
    """Converts an environment block as returned from CreateEnvironmentBlock to a Python dict of key/value strings.

    An environment block is a void pointer. The pointer points to a sequence of strings. Each string is terminated with a
    null character. The final string is terminated by an additional null character.
    """
    assert block.value is not None
    w_char_size = sizeof(c_wchar)
    cur: int = block.value
    key_val_str = cast(cur, c_wchar_p).value
    env: dict[str, str] = {}
    while key_val_str:
        key, val = key_val_str.split("=", maxsplit=1)
        env[key] = val
        # Advance pointer by string length plus a null character
        cur += (len(key_val_str) + 1) * w_char_size
        key_val_str = cast(cur, c_wchar_p).value

    return env


def environment_block_from_dict(env: dict[str, str]) -> c_wchar_p:
    """Converts a Python dictionary representation of an environment into a character buffer as expected by the
    lpEnvironment argument to the CreateProcess* family of win32 functions.

    Note: The returned c_char_p is pointing to the internal contents of an immutable python string; that is
        to say that it will be garbage collected, and the caller need not worry about deallocating it.
    """
    # Create a string with null-character delimiters between each "key=value" string
    null_delimited = "\0".join(f"{key}={value}" for key, value in env.items())
    # Add a final null-terminator character
    env_block_str = null_delimited + "\0"

    return c_wchar_p(env_block_str)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from __future__ import annotations

from typing import Any, MutableMapping, TypedDict
from enum import Flag, auto
import logging

__all__ = ["LOG", "LoggerAdapter", "LogExtraInfo", "LogContent"]


class LogContent(Flag):
    """
    A Flag Enum which describes the content of a log record.
    """

    BANNER = auto()  # Logs which contain a banner, such as to visually seperate log sections
    FILE_PATH = auto()  # Logs which contain a filepath
    FILE_CONTENTS = auto()  # Logs which contain the contents of a file
    COMMAND_OUTPUT = auto()  # Logs which contain the output of a command run
    EXCEPTION_INFO = (
        auto()
    )  # Logs which contain an exception openjd encountered, potentially including sensitive information such as filepaths or host information.
    PROCESS_CONTROL = (
        auto()
    )  # Logs which contain information related to starting, killing, or signalling processes.
    PARAMETER_INFO = (
        auto()
    )  # Logs which contain details about parameters and their values pertaining to the running action
    HOST_INFO = (
        auto()
    )  # Logs which contain details about the system environment, e.g. dependency versions, OS name, CPU architecture.


class LogExtraInfo(TypedDict):
    """
    A TypedDict which contains extra information to be added to the "extra" key of a log record.
    """

    openjd_log_content: LogContent | None


class LoggerAdapter(logging.LoggerAdapter):
    """
    LoggerAdapter which merges the "extra" kwarg instead of replacing with what the LoggerAdapter was initialized with.
    """

    def process(
        self, msg: Any, kwargs: MutableMapping[str, Any]
    ) -> tuple[Any, MutableMapping[str, Any]]:
        """
        Typically the LoggerAdaptor simply replaces the `extra` key in the kwargs with the one initialized with the
        adapter. However, we want to merge the two dictionaries, so we override it here.
        """
        if "extra" not in kwargs:
            kwargs["extra"] = self.extra
        else:
            kwargs["extra"] |= self.extra
        return msg, kwargs


# Name the logger for the sessions module, rather than this specific file
LOG = logging.getLogger(".".join(__name__.split(".")[:-1]))
"""
The logger of the openjd sessions module. The logger has the name openjd.sessions and is used 
throughout the openjd sessions module to provide information on actions the module is taking, as
well as any output from commands run during Sessions.

Some LogRecords sent to the logger will have an extra attribute named "openjd_log_content" who's 
value is a LogContent which provides information on what data is contained in the LogRecord, or None
if there is no applicable LogContent field (for example, a message like "Ending Session")

If the LogRecord does not have the "openjd_log_content" no guarantees are made as to what content
is in the LogRecord. LogRecords that contain LogContent.EXECPTION_INFO may also transitively include
potentially sensitive information like filepaths or host info due to the nature of exception messages.
"""
LOG.setLevel(logging.INFO)

_banner_log_extra = LogExtraInfo(openjd_log_content=LogContent.BANNER)


def log_section_banner(logger: LoggerAdapter, section_title: str) -> None:
    logger.info("")
    logger.info("==============================================", extra=_banner_log_extra)
    logger.info(f"--------- {section_title}", extra=_banner_log_extra)
    logger.info("==============================================", extra=_banner_log_extra)


def log_subsection_banner(logger: LoggerAdapter, section_title: str) -> None:
    logger.info("----------------------------------------------", extra=_banner_log_extra)
    logger.info(section_title, extra=_banner_log_extra)
    logger.info("----------------------------------------------", extra=_banner_log_extra)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

# Note: This script will only be run as a subprocess.

import sys
import ctypes
from ctypes.wintypes import BOOL, DWORD

# This assertion short-circuits mypy from type checking this module on platforms other than Windows
# https://mypy.readthedocs.io/en/stable/common_issues.html#python-version-and-system-platform-checks
assert sys.platform == "win32"

kernel32 = ctypes.WinDLL("kernel32")

kernel32.AllocConsole.restype = BOOL
kernel32.AllocConsole.argtypes = []

# https://learn.microsoft.com/en-us/windows/console/attachconsole
kernel32.AttachConsole.restype = BOOL
kernel32.AttachConsole.argtypes = [
    DWORD,  # [in] dwProcessId
]

# https://learn.microsoft.com/en-us/windows/console/freeconsole
kernel32.FreeConsole.restype = BOOL
kernel32.FreeConsole.argtypes = []

kernel32.GenerateConsoleCtrlEvent.restype = BOOL
kernel32.GenerateConsoleCtrlEvent.argtypes = [
    DWORD,  # [in] dwCtrlEvent
    DWORD,  # [in] dwProcessGroupId
]

CTRL_C_EVENT = 0
CTRL_BREAK_EVENT = 1

ATTACH_PARENT_PROCESS = -1


def signal_process(pgid: int):
    # Send signal can only target processes in the same console.
    # We first detach from the current console and re-attach to that of process group.
    if not kernel32.FreeConsole():
        raise ctypes.WinError()
    if not kernel32.AttachConsole(pgid):
        raise ctypes.WinError()

    # Send the signal
    # We send CTRL-BREAK as handler for it cannnot be disabled.
    # https://learn.microsoft.com/en-us/windows/console/ctrl-c-and-ctrl-break-signals

    # We only send CTRL-BREAK
    # if not kernel32.GenerateConsoleCtrlEvent(CTRL_C_EVENT, pgid):
    #     raise ctypes.WinError()
    if not kernel32.GenerateConsoleCtrlEvent(CTRL_BREAK_EVENT, pgid):
        raise ctypes.WinError()

    if not kernel32.FreeConsole():
        raise ctypes.WinError()
    if not kernel32.AttachConsole(ATTACH_PARENT_PROCESS):
        raise ctypes.WinError()


if __name__ == "__main__":
    signal_process(int(sys.argv[1]))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import json
import os
import stat
import sys
import uuid
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
from logging import Filter
from os import name as os_name
from os import stat as os_stat
from pathlib import Path
from tempfile import mkstemp
from types import TracebackType
from typing import TYPE_CHECKING, Any, Callable, Optional, Type, Union

from openjd.model import (
    JobParameterValues,
    ParameterValue,
    ParameterValueType,
    SpecificationRevision,
    SymbolTable,
    TaskParameterSet,
)
from openjd.model import version as model_version
from openjd.model.v2023_09 import (
    ValueReferenceConstants as ValueReferenceConstants_2023_09,
)
from ._action_filter import ActionMessageKind, ActionMonitoringFilter
from ._embedded_files import write_file_for_user
from ._logging import LOG, log_section_banner, LoggerAdapter, LogExtraInfo, LogContent
from ._os_checker import is_posix, is_windows
from ._path_mapping import PathMappingRule
from ._runner_base import ScriptRunnerBase
from ._runner_env_script import EnvironmentScriptRunner
from ._runner_step_script import StepScriptRunner
from ._session_user import SessionUser
from ._subprocess import LoggingSubprocess
from ._tempdir import TempDir, custom_gettempdir
from ._types import (
    ActionState,
    EnvironmentIdentifier,
    EnvironmentModel,
    StepScriptModel,
)
from ._version import version

if TYPE_CHECKING:
    from openjd.model.v2023_09._model import EnvironmentVariableObject

__all__ = ("SessionState", "Session", "EnvironmentIdentifier")


class SessionState(str, Enum):
    READY = "ready"
    """The state of a Session when it is ready to run actions.
    """

    RUNNING = "running"
    """The state of a Session while it is actively running an action.
    """

    CANCELING = "canceling"
    """The state of a Session that is in the process of canceling the currently
    running action.
    """

    READY_ENDING = "ready_ending"
    """The state of a Session when it is ready to run only Environment End actions.
    The Session has previously experienced an error or cancelation running
    one of its actions, and is now only allowed to run Environment End actions to
    clean up the Session context."""

    ENDED = "ended"
    """Terminal state of a Session that has ended and can no longer run any actions."""


@dataclass(frozen=True)
class ActionStatus:
    state: ActionState
    """The runtime state of the action."""

    progress: Optional[float] = None
    """The progress of the action as reported by an "openjd_progress:" message"""

    status_message: Optional[str] = None
    """The status message for the action as reported by an "openjd_status:" message"""

    fail_message: Optional[str] = None
    """The failure reason of the action as reported by an "openjd_fail:" message."""

    exit_code: Optional[int] = None
    """The exit code of the action's process, if it has exited.
    Note: This may be None in SUCCESS & FAILED states.
    e.g.
        SUCCESS - Entered an environment that ran no Action to enter it.
        FAILED - Failed before trying to run the Action subprocess, such as
            failing to write embedded files to disk.
    """


@dataclass
class EnvironmentVariableSetChange:
    name: str
    value: str


@dataclass
class EnvironmentVariableUnsetChange:
    name: str


EnvironmentVariableChange = Union[EnvironmentVariableSetChange, EnvironmentVariableUnsetChange]


class SimplifiedEnvironmentVariableChanges:
    """
    Keeps track of what variables need to be set and unset for an environment. On Windows all environment keys
    are stored as uppercase. This is because while Windows environment variables are case insensitive, the win32 api
    allows you to store multiple keys of mixed case, for example the win32 api would allow you
    to set both "PATH" and "Path" as environment variable keys. This leads to undefined behaviour when calling one of the keys.
    """

    def __init__(self, initial_variables: Union[dict[str, str], "EnvironmentVariableObject"]):
        self._to_set: dict[str, Optional[str]]

        if is_windows():
            self._to_set = {}
            for var_name, var_value in initial_variables.items():
                self._to_set[var_name.upper()] = var_value
        else:
            self._to_set = dict(initial_variables)  # Make a copy

    def simplify_ordered_changes(self, changes: list[EnvironmentVariableChange]) -> None:
        """Apply a given list of sets and unsets to the current state in order"""
        for change in changes:
            if isinstance(change, EnvironmentVariableSetChange):
                self._to_set[change.name.upper() if is_windows() else change.name] = change.value
            elif isinstance(change, EnvironmentVariableUnsetChange):
                self._to_set[change.name.upper() if is_windows() else change.name] = None
            else:
                raise ValueError("Unknown type of environment variable change.")

    def apply_to_environment(self, env_vars: dict[str, Optional[str]]) -> None:
        """Modify a given dictionary of environment variables to reflect the changes"""
        if is_windows():
            # Environment variables on Windows are case insensitive when used, but are case sensitive when
            # set via the Windows API.
            for var_name, var_value in env_vars.copy().items():
                del env_vars[var_name]
                env_vars[var_name.upper()] = var_value

        # Note: An env var value of None means to unset that variable
        env_vars.update(self._to_set)


SessionCallbackType = Callable[[str, ActionStatus], None]


class Session(object):
    """A context for running actions of an Open Job Description Job.

    In Open Job Description, the Tasks for a Job's Steps are run within the context of a *Session*.
    Each Step in a Job defines the properties of the Session that are required to
    run its Tasks. Open Job Description sessions enable users to amortize expensive or time-consuming
    setup and tear-down operations in the worker's environment before and after a sequence
    of Tasks.

    A Session starts 0 or more *Environments* in the order given on the worker when
    it is started, and ends those Environments in reverse order when the Session is
    no longer needed. Each Environment defines a start and end *Action* — a command/script
    defined by the end-user — that is run on the worker when starting or ending the
    Environment.  The actions for these environments, as with all actions, are each
    run in their own operating system process.

    All stdout and stderr from the subproceseses run within this Session, and any additional
    logging generated by the Session itself, are forwarded to the Open Job Description sessions
    module's LOG Logger at log level INFO. The LogRecords sent to the log have an
    extra attribute named "session_id" whose value is the session_id that was passed
    to the constructor of the Session.

    Each Session has its own temporary working directory, referred to as the Session
    Working Directory. Instantiating this class creates this directory for the duration
    of the Session, and you are expected to call the cleanup() method on your Session
    to delete that directory when done (this is done automatically if using the Session
    as a context manager).
    On POSIX:
    - The Session Working Directory's owner is the process owner.
    - If a PosixSessionUser is provided, then the group of that user is the group owner
      of the the directory.
    On Windows:
    - All files and directories within the Session Working Directory inherit the
      directory's ACL.
    - The Session Working Directory's ACL is set so that the process owner has full
      control
    - If a WindowsSessionUser is provided then the user within that SessionUser
      is also given Modify access.
    """

    _state: SessionState

    _session_id: str
    """The application-provided id for this Session.
    """

    _logger: LoggerAdapter
    """The logger that all of this Session's running processes will send their logs to.
    """

    _ending_only: bool
    """The Session has previously experienced an error or cancelation of a running
    action, and can only run Environment-end actions.
    """

    _environments: dict[
        EnvironmentIdentifier,
        EnvironmentModel,
    ]
    """A mapping of identifier to Environment for each Environment entered in the session.
    """

    _environments_entered: list[EnvironmentIdentifier]
    """A list of the Environments entered (either successfully or unsuccessfully[failed/canceled]),
    in the order that they were entered.
    Environments must be exited in the reverse order to that which they were entered.
    """

    _runner: Optional[ScriptRunnerBase]
    """The currently running runner, if there is one.
    """

    _running_environment_identifier: Optional[str]
    """If we're running an environment action then this will be the
    identifier of that environment; otherwise it will be None.
    """

    _process_env: dict[str, str]
    """Mapping of environment variable names to values. Used as the shell/os environment
    when running a subprocess.
    """

    _created_env_vars: dict[EnvironmentIdentifier, SimplifiedEnvironmentVariableChanges]
    """OS environment variables defined by Open Job Description Environments
    """

    _log_filter: Filter
    """The handler that we've hooked to the LOG. Removed when the Session is deleted.
    """

    _working_dir: Optional[TempDir] = None
    """The Session Working Directory.
    """

    _files_dir: TempDir
    """The subdirectory of the Working Directory where embedded files
    are materialized.
    """

    _retain_working_dir: bool
    """If True, then the working directory is not deleted on cleanup.
    """

    _user: Optional[SessionUser]
    """The specific OS user to run subprocesses as, and whom will have permissions
    to the Session's Working Directory.
    Defaults to the current process user.
    """

    _job_parameter_values: JobParameterValues
    """Values for any defined Job Parameters.
    This is a dictionary.
        key = Parameter name (e.g. "Foo")
        value = Parameter's type and value
    """

    _cleanup_called: bool
    """Whether or not the application has called cleanup.
    If not, then we will call it automatically in __del__.
    """

    _callback: Optional[SessionCallbackType]
    """If provided, then this callback will be invoked on every:
        1. Open Job Description action message in the log (lines that start with openjd_)
        2. Completion/exit of the current action.
    The callback takes two arguments:
        - The session_id of this session; and
        - The updated value of Session.action_status; this contains the
            runtime state of the Action (running, etc) and optionally the
            progress, exit_code, and any output messages.
    """

    _path_mapping_rules: Optional[list[PathMappingRule]]
    """A list of the Path Mapping rules to communicate to Actions run via this Session.
    """

    _session_root_directory: Optional[Path]
    """If non-None, then this is the directory within-which the Session creator wants
    the session's working directory to be created.
    """

    # Status fields for the currently running process, if any.
    _action_state: Optional[ActionState]
    _action_progress: Optional[float]
    _action_status_message: Optional[str]
    _action_fail_message: Optional[str]
    _action_exit_code: Optional[int]

    def __init__(
        self,
        *,
        session_id: str,
        job_parameter_values: JobParameterValues,
        path_mapping_rules: Optional[list[PathMappingRule]] = None,
        retain_working_dir: bool = False,
        user: Optional[SessionUser] = None,
        callback: Optional[SessionCallbackType] = None,
        os_env_vars: Optional[dict[str, str]] = None,
        session_root_directory: Optional[Path] = None,
    ):
        """
        Arguments:
            session_id (str): An application-defined string value with which the session is identified.
            job_parameter_values (JobParameterValues): Values for any defined Job Parameters. This is a
                dictionary where the keys are parameter names, and the values are instances of
                ParameterValue (a dataclass containing the type and value of the parameter)
            path_mapping_rules (Optional[list[PathMappingRule]]): A list of the path mapping rules to apply
                within all actions running within this session. Defaults to None.
            retain_working_dir (bool, optional): If set, then the Session's Working Directory
                is not deleted when this Session object is deleted. Defaults to False.
            user (Optional[SessionUser]): The specific OS user to run subprocesses as, and whom
                will have permissions to the Session's Working Directory.
                Defaults to the current process user.
            callback (Optional[SessionCallbackType]): If provided, then this callback will be
                invoked on every:
                    1. Start of an Action (when the subprocess is actually running);
                    2. Open Job Description action message in the log (lines that start with openjd_)
                    3. Completion/exit of the current action.
                The callback takes two arguments:
                    - The session_id of this session; and
                    - The updated value of Session.action_status; this contains the
                        runtime state of the Action (running, etc) and optionally the
                        progress, exit_code, and any output messages.
                An implementation that uses this callback must ensure that the callback exits
                very rapidly; not doing so will delay processing of stdout/stderr of running
                subprocesses.
                WARNING: This callback may be called from the same thread that initiated
                a Script run in this session. This may happen if an error was encountered
                prior to trying to run the Action (e.g. failing to write an embedded file to
                disk); if the Action runs very quickly; or other similar circumstances.
            os_env_vars (Optional[dict[str,str]]): Definitions for additional OS Environment
                Variables that should be injected into all running processes in the Session.
                    Key: Environment variable name
                    Value: Value for the environment variable.
            session_root_directory (Optional[Path]): If provided, then:
                1. The given directory must already exist;
                2. The 'user' (if given) must have at least read permissions to it; and
                3. The Working Directory for this Session will be created in the given directory.
                If not provided, then the default of gettempdir()/"openjd" is used instead.

        Raises:
            RuntimeError - If the Session initialization fails for any reason.
        """

        self._session_id = session_id
        self._ending_only = False
        self._environments = dict()
        self._environments_entered = list()
        self._runner = None
        self._running_environment_identifier = None
        self._process_env = dict(os_env_vars) if os_env_vars else dict()
        self._created_env_vars = dict()
        self._retain_working_dir = retain_working_dir
        self._user = user
        self._job_parameter_values = dict(job_parameter_values) if job_parameter_values else dict()
        self._cleanup_called = False
        self._callback = callback
        self._path_mapping_rules = path_mapping_rules[:] if path_mapping_rules else None
        if self._path_mapping_rules is not None:
            # Path mapping rules are applied in order of longest to shortest source path,
            # so sort them for when we apply them.
            self._path_mapping_rules.sort(key=lambda rule: -len(rule.source_path.parts))
        self._session_root_directory = session_root_directory
        if self._session_root_directory is not None:
            if not self._session_root_directory.is_dir():
                raise RuntimeError(
                    f"Ensure that the root directory ({str(self._session_root_directory)}) exists and is a directory."
                )
        self._reset_action_state()

        # Set up our logging hook & callback
        self._log_filter = ActionMonitoringFilter(
            session_id=self._session_id, callback=self._action_log_filter_callback
        )
        LOG.addFilter(self._log_filter)
        self._logger = LoggerAdapter(LOG, extra={"session_id": self._session_id})

        host_info_extra = LogExtraInfo(openjd_log_content=LogContent.HOST_INFO)
        self._logger.info(f"openjd.model Library Version: {model_version}", extra=host_info_extra)
        self._logger.info(f"openjd.sessions Library Version: {version}", extra=host_info_extra)
        self._logger.info(
            "Installed at: %s", str(Path(__file__).resolve().parent.parent), extra=host_info_extra
        )
        self._logger.info(f"Python Interpreter: {sys.executable}", extra=host_info_extra)
        self._logger.info(
            "Python Version: %s", sys.version.replace("\n", " - "), extra=host_info_extra
        )
        self._logger.info(f"Platform: {sys.platform}", extra=host_info_extra)
        self._logger.info(f"Initializing Open Job Description Session: {self._session_id}")

        try:
            self._working_dir = self._create_working_directory()
            self._files_dir = self._create_files_directory()
        except RuntimeError as exc:
            self._logger.error(
                f"ERROR creating Session Working Directory: {str(exc)}",
                extra=LogExtraInfo(
                    openjd_log_content=LogContent.COMMAND_OUTPUT | LogContent.FILE_PATH
                ),
            )
            self._state = SessionState.ENDED
            raise

        self._logger.info(
            f"Session Working Directory: {str(self.working_directory)}",
            extra=LogExtraInfo(openjd_log_content=LogContent.FILE_PATH),
        )
        self._logger.info(
            f"Session's Embedded Files Directory: {str(self.files_directory)}",
            extra=LogExtraInfo(openjd_log_content=LogContent.FILE_PATH),
        )

        self._state = SessionState.READY

    def cleanup(self) -> None:
        """Cleanup all resources created by this Session"""
        if self._cleanup_called:
            return
        self._cleanup_called = True
        if self._working_dir is not None and not self._retain_working_dir:
            log_section_banner(self._logger, "Session Cleanup")
            self._logger.info(
                f"Deleting working directory: {str(self.working_directory)}",
                extra=LogExtraInfo(openjd_log_content=LogContent.FILE_PATH),
            )
            try:
                # If running as a different user, then that user could have written files to the
                # session diretory that make removing it as our user impossible. So, do a 2-phase
                # removal: 1/ `sudo -u <user> -i rm -rf <sessiondir>`, and then 2/ doing a normal
                # recursive removal to delete the stuff that only this user can delete.
                if self._user is not None:
                    files = [str(f) for f in self.working_directory.glob("*")]

                    if is_posix():
                        recursive_delete_cmd = ["rm", "-rf"]
                    else:
                        recursive_delete_cmd = [
                            "start",
                            '"Powershell"',
                            "/high",
                            "/wait",
                            "/b",
                            "powershell",
                            "-Command",
                            "Remove-Item",
                            "-Recurse",
                            "-Force",
                        ]
                        files = [", ".join(files)]

                    subprocess = LoggingSubprocess(
                        logger=self._logger,
                        args=recursive_delete_cmd + files,
                        user=self._user,
                    )
                    # Note: Blocking call until the process has exited
                    subprocess.run()

                self._working_dir.cleanup()
            except RuntimeError as exc:
                # Warn if we couldn't cleanup the temporary files for some reason.
                self._logger.exception(
                    exc,
                    extra=LogExtraInfo(
                        openjd_log_content=LogContent.EXCEPTION_INFO | LogContent.FILE_PATH
                    ),
                )

        LOG.removeFilter(self._log_filter)
        del self._log_filter
        if self._runner:
            self._runner.shutdown()
            self._runner = None
        self._state = SessionState.ENDED

    def __enter__(self) -> "Session":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.cleanup()

    # ========================
    #  Properties

    @property
    def working_directory(self) -> Path:
        """The directory that was created for this Session's working files.
        This is available in a Job Template's format string expressions as
        Session.WorkingDirectory
        """
        assert self._working_dir is not None
        return self._working_dir.path

    @property
    def files_directory(self) -> Path:
        """The subdirectory of the working_directory where files that have
        been inlined into a Job Template are stored.
        """
        return self._files_dir.path

    @property
    def state(self) -> SessionState:
        """Fetch the current state of this Session."""
        return self._state

    @property
    def action_status(self) -> Optional[ActionStatus]:
        """Obtain the status of the currently running, or previously running, Action
        Includes progress, status messages, and exit code; if they are available.
        """
        # If we don't have an action state, then we're not running or haven't run
        # anything yet.
        if self._action_state is None:
            return None
        return ActionStatus(
            state=self._action_state,
            progress=self._action_progress,
            status_message=self._action_status_message,
            fail_message=self._action_fail_message,
            exit_code=self._action_exit_code,
        )

    @property
    def environments_entered(self) -> tuple[EnvironmentIdentifier, ...]:
        """Returns an immutable list of the identifiers for Environments
        that have been entered, in the order that that have been entered.
        """
        return tuple(self._environments_entered)

    # =========================
    #  Running Actions

    def cancel_action(
        self, *, time_limit: Optional[timedelta] = None, mark_action_failed=False
    ) -> None:
        """Initiate a cancelation of the currently running Script if there is one.

        Arguments:
            time_limit (Optional[timedelta]): If provided, then the cancel must be completed within
                the given number of seconds. This overrides the current action's cancelation method
                and is intended for urgent cancels (e.g. in response to the controlling process
                getting a SIGTERM). Note: a value of 0 turns a notify-then-terminate cancel into a
                terminate

        Raises:
            RuntimeError: When there is no Script running.
        """
        if self.state != SessionState.RUNNING:
            raise RuntimeError("No actions are running")
        # For the type checker
        assert self._runner is not None

        self._runner.cancel(time_limit=time_limit, mark_action_failed=mark_action_failed)

    def enter_environment(
        self,
        *,
        environment: EnvironmentModel,
        identifier: Optional[EnvironmentIdentifier] = None,
        os_env_vars: Optional[dict[str, str]] = None,
    ) -> EnvironmentIdentifier:
        """Enters an Open Job Description Environment within this Session.
        This method is non-blocking; it will exit when the subprocess is either confirmed to have
        started running, or has failed to be started.

        Arguments:
            environment (EnvironmentScriptModel): An Environment Script from a supported
                Open Job Description schema version. This Script should not have any of its Format Strings
                evaluated before hand.
            identifier (Optional[EnvironmentIdentifier]): If provided then this is the identifier
                that the Environment will be known by to this Session.
                Default: An identifier is randomly generated.
            os_env_vars (Optional[dict[str,str]): Definitions for additional OS Environment
                Variables that should be injected into the process that is run for this action.
                Values provided override values provided to the Session constructor, and are overriden
                by values defined in Environments.
                    Key: Environment variable name
                    Value: Value for the environment variable.

        Returns:
            EnvironmentIdentifier: An identifier by which the Environment is known by to this Session.
                Pass this identifier to exit_environment() when exiting this Environment.
        """
        if self.state != SessionState.READY:
            raise RuntimeError("Session must be in the READY state to enter an Environment.")
        if identifier is not None and identifier in self._environments:
            raise RuntimeError(
                f"Environment {identifier} has already been entered in this Session."
            )
        log_section_banner(self._logger, f"Entering Environment: {environment.name}")

        self._reset_action_state()

        if identifier is None:
            identifier = f"{self._session_id}:{uuid.uuid4().hex}"

        self._environments[identifier] = environment
        self._environments_entered.append(identifier)
        self._running_environment_identifier = identifier

        symtab = self._symbol_table(environment.revision)

        if environment.variables is not None:
            # We must process the current environment's variables
            # before we call _evaluate_current_session_env_vars()
            # otherwise, we will end up running onEnter without
            # the environment variables of the current environment
            # being set.
            resolved_variables = self._resolve_env_variable_format_strings(
                symtab, environment.variables
            )
            for name, value in resolved_variables.items():
                self._logger.info(
                    "Setting: %s=%s",
                    name,
                    value,
                    extra=LogExtraInfo(openjd_log_content=LogContent.PARAMETER_INFO),
                )
            env_var_changes = SimplifiedEnvironmentVariableChanges(resolved_variables)
            self._created_env_vars[identifier] = env_var_changes
        else:
            # Running the environment may define environment variable
            # mutations via its stdout. We create an empty env changes
            # object to capture these.
            self._created_env_vars[identifier] = SimplifiedEnvironmentVariableChanges(
                dict[str, str]()
            )

        # Must be called _after_ we append to _environments_entered
        action_env_vars = self._evaluate_current_session_env_vars(os_env_vars)

        self._materialize_path_mapping(environment.revision, action_env_vars, symtab)

        # Sets the subprocess running.
        # Returns immediately after it has started, or is running
        self._action_state = ActionState.RUNNING
        self._state = SessionState.RUNNING
        # Note: This may fail immediately (e.g. if we cannot write embedded files to disk),
        # so it's important to set the action_state to RUNNING before calling enter(), rather
        # than after -- enter() itself may end up setting the action state to FAILED.

        self._runner = EnvironmentScriptRunner(
            logger=self._logger,
            user=self._user,
            os_env_vars=action_env_vars,
            session_working_directory=self.working_directory,
            startup_directory=self.working_directory,
            callback=self._action_callback,
            environment_script=environment.script,
            symtab=symtab,
            session_files_directory=self.files_directory,
        )
        self._runner.enter()

        return identifier

    def exit_environment(
        self, *, identifier: EnvironmentIdentifier, os_env_vars: Optional[dict[str, str]] = None
    ) -> None:
        """Exits an Open Job Description Environment from this Session.
        This method is non-blocking; it will exit when the subprocess is either confirmed to have
        started running, or has failed to be started.

        Note that Environments *MUST* be exited in the opposite order in which they were entered.
        It is an error to do otherwise.

        Arguments:
            identifier (EnvironmentIdentifier): The identifier of the previously entered
                Environment to exit.
            os_env_vars (Optional[dict[str,str]): Definitions for additional OS Environment
                Variables that should be injected into the process that is run for this action.
                Values provided override values provided to the Session constructor, and are overriden
                by values defined in Environments.
                    Key: Environment variable name
                    Value: Value for the environment variable.

        Raises:
            ValueError - If the given identifier is not that of the next one that must be exited.
        """
        if self.state != SessionState.READY and self.state != SessionState.READY_ENDING:
            raise RuntimeError(
                "Session must be in the READY or READY_ENDING state to exit an Environment."
            )
        if identifier not in self._environments:
            raise RuntimeError(f"Cannot exit unknown Environment with identifier {identifier}")
        if self._environments_entered[-1] != identifier:
            raise RuntimeError(
                f"Cannot exit Environment {identifier}. Must exit Environment {self._environments_entered[-1]} first."
            )

        self._reset_action_state()

        # Once we've started exiting environments, then we can only exit environments.
        self._ending_only = True

        environment = self._environments[identifier]
        log_section_banner(self._logger, f"Exiting Environment: {environment.name}")

        # Must be run _before_ we pop _environments_entered
        action_env_vars = self._evaluate_current_session_env_vars(os_env_vars)

        # Remove the environment from our tracking since we're now exiting it.
        del self._environments[identifier]
        self._environments_entered.pop()

        self._running_environment_identifier = identifier

        symtab = self._symbol_table(environment.revision)
        self._materialize_path_mapping(environment.revision, action_env_vars, symtab)
        # Sets the subprocess running.
        # Returns immediately after it has started, or is running
        self._action_state = ActionState.RUNNING
        self._state = SessionState.RUNNING
        # Note: This may fail immediately (e.g. if we cannot write embedded files to disk),
        # so it's important to set the action_state to RUNNING before calling exit(), rather
        # than after -- exit() itself may end up setting the action state to FAILED.

        self._runner = EnvironmentScriptRunner(
            logger=self._logger,
            user=self._user,
            os_env_vars=action_env_vars,
            session_working_directory=self.working_directory,
            startup_directory=self.working_directory,
            callback=self._action_callback,
            environment_script=environment.script,
            symtab=symtab,
            session_files_directory=self.files_directory,
        )
        self._runner.exit()

    def run_task(
        self,
        *,
        step_script: StepScriptModel,
        task_parameter_values: TaskParameterSet,
        os_env_vars: Optional[dict[str, str]] = None,
        log_task_banner: bool = True,
    ) -> None:
        """Run a Task within the Session.
        This method is non-blocking; it will exit when the subprocess is either confirmed to have
        started running, or has failed to be started.

        Arguments:
            step_script (StepScriptModel): The Step Script that the Task will be running.
            task_parameter_values (TaskParameterSet): Values of the Task parameters that define the
                specific Task. This is a dictionary where the keys are parameter names, and the values
                are instances of ParameterValue (a dataclass containing the type and value of the parameter)
            os_env_vars (Optional[dict[str,str]): Definitions for additional OS Environment
                Variables that should be injected into the process that is run for this action.
                Values provided override values provided to the Session constructor, and are overriden
                by values defined in Environments.
                    Key: Environment variable name
                    Value: Value for the environment variable.
            log_task_banner (bool): Whether to log a banner before running the Task.
                Default: True
        """
        if self.state != SessionState.READY:
            raise RuntimeError("Session must be in the READY state to run a task.")

        if log_task_banner:
            log_section_banner(self._logger, "Running Task")

        if task_parameter_values:
            self._logger.info(
                "Parameter values:",
                extra=LogExtraInfo(openjd_log_content=LogContent.PARAMETER_INFO),
            )
            for name, value in task_parameter_values.items():
                self._logger.info(
                    f"{name}({str(value.type.value)}) = {value.value}",
                    extra=LogExtraInfo(openjd_log_content=LogContent.PARAMETER_INFO),
                )

        self._reset_action_state()
        symtab = self._symbol_table(step_script.revision, task_parameter_values)
        action_env_vars = self._evaluate_current_session_env_vars(os_env_vars)
        self._materialize_path_mapping(step_script.revision, action_env_vars, symtab)
        self._runner = StepScriptRunner(
            logger=self._logger,
            user=self._user,
            os_env_vars=action_env_vars,
            session_working_directory=self.working_directory,
            startup_directory=self.working_directory,
            callback=self._action_callback,
            script=step_script,
            symtab=symtab,
            session_files_directory=self.files_directory,
        )
        # Sets the subprocess running.
        # Returns immediately after it has started, or is running
        self._action_state = ActionState.RUNNING
        self._state = SessionState.RUNNING
        # Note: This may fail immediately (e.g. if we cannot write embedded files to disk),
        # so it's important to set the action_state to RUNNING before calling run(), rather
        # than after -- run() itself may end up setting the action state to FAILED.
        self._runner.run()

    # =========================
    #  Helpers

    def _reset_action_state(self) -> None:
        """Reset the internal action state.
        This resets to a state equivalent to having nothing running.
        """
        self._action_state = None
        self._action_progress = None
        self._action_status_message = None
        self._action_fail_message = None
        self._action_exit_code = None
        self._running_environment_identifier = None
        if self._runner:
            self._runner.shutdown()
            self._runner = None

    def _symbol_table(
        self,
        version: SpecificationRevision,
        task_parameter_values: Optional[TaskParameterSet] = None,
    ) -> SymbolTable:
        """Construct a SymbolTable, with fully qualified value names, suitable for running a Script."""

        def processed_parameter_value(param: ParameterValue) -> str:
            if param.type == ParameterValueType.PATH and self._path_mapping_rules is not None:
                # Apply path mapping rules in the order given until one does a replacement
                for rule in self._path_mapping_rules:
                    changed, result = rule.apply(path=param.value)
                    if changed:
                        return result
            return param.value

        if version == SpecificationRevision.v2023_09:
            symtab = SymbolTable()
            symtab[ValueReferenceConstants_2023_09.WORKING_DIRECTORY.value] = str(
                self.working_directory
            )
            for param_name, param_props in self._job_parameter_values.items():
                symtab[
                    f"{ValueReferenceConstants_2023_09.JOB_PARAMETER_RAWPREFIX.value}.{param_name}"
                ] = param_props.value
                symtab[
                    f"{ValueReferenceConstants_2023_09.JOB_PARAMETER_PREFIX.value}.{param_name}"
                ] = processed_parameter_value(param_props)
            if task_parameter_values:
                for param_name, param_props in task_parameter_values.items():
                    symtab[
                        f"{ValueReferenceConstants_2023_09.TASK_PARAMETER_RAWPREFIX.value}.{param_name}"
                    ] = param_props.value
                    symtab[
                        f"{ValueReferenceConstants_2023_09.TASK_PARAMETER_PREFIX.value}.{param_name}"
                    ] = processed_parameter_value(param_props)
            return symtab
        else:
            raise NotImplementedError(f"Schema version {str(version.value)} is not supported.")

    def _openjd_session_root_dir(self) -> Path:
        """
        Returns (and creates if necessary) the top-level directory where Open Job Description step session
        directories are kept
        """
        if self._session_root_directory is not None:
            return self._session_root_directory

        tempdir = Path(custom_gettempdir(self._logger))

        # Note: If this doesn't have group permissions, then we will be unable to access files
        #  under this directory if the default group of the current user is the group that
        #  is shared with a job user. The group permissions override the world permissions
        #  when the accessor is in the group.
        # 0o755
        mode = stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH
        tempdir.mkdir(exist_ok=True, mode=mode)
        # tempdir might already exist with the incorrect permissions, so set the permissions.
        os.chmod(tempdir, mode=mode)
        return tempdir

    def _create_working_directory(self) -> TempDir:
        """Creates and returns the temporary working directory for this Session"""
        root_dir = self._openjd_session_root_dir()

        if os_name == "posix":
            # Check the sticky bit. If we have any world-writeable parents to
            # the root_dir that don't have the sticky bit set, then
            # the system has an insecure setup for multiuser systems.
            for parent in root_dir.parents:
                parent_stat = os_stat(parent)
                # Note There is a nuanced security risks to putting the session directory in a world-writable parent
                # directory. Normally, users with write permissions to a directory can delete files/directories within
                # that directory and this is a problem for world-writable dirs like /tmp. Linux distros typically
                # default to the system temp dir having the sticky bit set which restricts deletion of files/dirs in
                # world-writable dirs to only the owning user or a privileged/root user. Not all distros may respect this,
                # or system administrators may unset the sticky bit.
                if (parent_stat.st_mode & stat.S_IWOTH) != 0 and (
                    parent_stat.st_mode & stat.S_ISVTX
                ) == 0:
                    self._logger.warning(
                        f"Sticky bit is not set on {str(parent)}. This may pose a risk when running work on this host as users may modify or delete files in this directory which do not belong to them.",
                        extra=LogExtraInfo(
                            openjd_log_content=LogContent.HOST_INFO | LogContent.FILE_PATH
                        ),
                    )

        # Raises: RuntimeError
        return TempDir(dir=root_dir, prefix=self._session_id, user=self._user, logger=self._logger)

    def _create_files_directory(self) -> TempDir:
        """Creates the subdirectory of the working directory in which we'll materialize
        any embedded files from the Job Template."""
        # Raises: RuntimeError
        return TempDir(
            dir=self.working_directory,
            prefix="embedded_files",
            user=self._user,
            logger=self._logger,
        )

    def _materialize_path_mapping(
        self, version: SpecificationRevision, os_env: dict[str, Optional[str]], symtab: SymbolTable
    ) -> None:
        """Materialize path mapping rules to disk and the os environment variables."""
        if self._path_mapping_rules:
            rules_dict = {
                "version": "pathmapping-1.0",
                "path_mapping_rules": [
                    {
                        "source_path_format": rule.source_path_format.value,
                        "source_path": str(rule.source_path),
                        "destination_path": str(rule.destination_path),
                    }
                    for rule in self._path_mapping_rules
                ],
            }
            symtab[ValueReferenceConstants_2023_09.HAS_PATH_MAPPING_RULES.value] = "true"
        else:
            rules_dict = dict()
            symtab[ValueReferenceConstants_2023_09.HAS_PATH_MAPPING_RULES.value] = "false"
        rules_json = json.dumps(rules_dict)
        file_handle, filename = mkstemp(dir=self.working_directory, suffix=".json", text=True)
        os.close(file_handle)
        write_file_for_user(Path(filename), rules_json, self._user)
        symtab[ValueReferenceConstants_2023_09.PATH_MAPPING_RULES_FILE.value] = str(filename)

    def _resolve_env_variable_format_strings(
        self, symtab: SymbolTable, variables: "EnvironmentVariableObject"
    ) -> dict[str, str]:
        """When definining an environment variable via an Environment entity's "variables" declaration,
        the values of those variables are format strings that must be evaluated. Do that, and return the
        result.
        """
        result = dict()
        for name, value in variables.items():
            result[name] = value.resolve(symtab=symtab)

        return result

    def _action_log_filter_callback(
        self, kind: ActionMessageKind, value: Any, cancel_action_mark_failed: bool = False
    ) -> None:
        """This callback is invoked by the ActionMonitoringFilter that we've attached to the LOG.
        This will be called whenever an "openjd" message is detected in the log stream.
        This will be invoked by the main thread in LoggingSubprocess that is forwarding
        all stdout/stderr to the logs. Delays here delay that main loop from processing
        output.
        """
        if kind == ActionMessageKind.PROGRESS:
            # Assert for the type checker; the type is guaranteed by the ActionMonitoringFilter
            assert isinstance(value, float)
            self._action_progress = value
        elif kind == ActionMessageKind.STATUS:
            # Assert for the type checker; the type is guaranteed by the ActionMonitoringFilter
            assert isinstance(value, str)
            self._action_status_message = value
        elif kind == ActionMessageKind.FAIL:
            # Assert for the type checker; the type is guaranteed by the ActionMonitoringFilter
            assert isinstance(value, str)
            self._action_fail_message = value

            if cancel_action_mark_failed:
                # Cancel the action and pass the failure message
                self.cancel_action(mark_action_failed=True)

        elif kind == ActionMessageKind.ENV:
            if self._running_environment_identifier is None:
                # Ignore the message if we're not running an environment
                return
            if cancel_action_mark_failed:
                # Assert for the type checker; the type is guaranteed by the ActionMonitoringFilter
                assert isinstance(value, str)

                # Cancel the action and pass the failure message
                self.cancel_action(mark_action_failed=True)
                self._action_fail_message = value

                return
            # Assert for the type checker; the type is guaranteed by the ActionMonitoringFilter
            assert isinstance(value, dict)
            # value = { "name": <name>, "value": <value> }
            env_vars = self._created_env_vars[self._running_environment_identifier]
            env_vars.simplify_ordered_changes(
                changes=[EnvironmentVariableSetChange(name=value["name"], value=value["value"])]
            )
            return
        elif kind == ActionMessageKind.UNSET_ENV:
            if self._running_environment_identifier is None:
                # Ignore the message if we're not running an environment
                return

            if cancel_action_mark_failed:
                # Assert for the type checker; the type is guaranteed by the ActionMonitoringFilter
                assert isinstance(value, str)

                # Cancel the action and pass the failure message
                self.cancel_action(mark_action_failed=True)
                self._action_fail_message = value

                return

            # Assert for the type checker; the type is guaranteed by the ActionMonitoringFilter
            assert isinstance(value, str)
            env_vars = self._created_env_vars[self._running_environment_identifier]
            env_vars.simplify_ordered_changes(changes=[EnvironmentVariableUnsetChange(name=value)])
            return
        else:  # ActionMessageKind.SESSION_RUNTIME_LOGLEVEL
            assert isinstance(value, int)
            self._logger.setLevel(value)
            return

        if self._callback:
            action_status = self.action_status
            # for the type checker
            assert action_status is not None
            self._callback(self._session_id, action_status)

    def _action_callback(self, state: ActionState) -> None:
        """This callback is invoked:
        1. When the Action process is successfully started, by the same thread that is running the
           process (so, this holds up IO processing);
        2. *After* the "run future" in ScriptRunnerBase has exited, by the same thread that was
           running that "run future";
        3. If we failed *before* actually running the Action (e.g. while trying to write embedded
           files to disk) then this will be invoked by the same thread that called Session.environment_*()
           or Session.run_task().

        We can be certain that the process is no longer running when this is called.
        """
        # For the type checker
        assert self._runner is not None

        self._action_exit_code = self._runner.exit_code
        self._action_state = state

        if state != ActionState.RUNNING:
            # Decide which between-action state to enter.
            if self._ending_only or self._action_state != ActionState.SUCCESS:
                # Sessions are "brittle". If there's a Task cancel or Failure then
                # we can only exit the Session.
                self._state = SessionState.READY_ENDING
            else:
                self._state = SessionState.READY

        if self._callback:
            action_status = self.action_status
            # for the type checker
            assert action_status is not None
            self._callback(self._session_id, action_status)

    def _evaluate_current_session_env_vars(
        self, extra_env_vars: Optional[dict[str, str]] = None
    ) -> dict[str, Optional[str]]:
        """Get a dictionary representing the cummulative state of env vars set
        and unset from the currently applied environments.
        """
        result = dict[str, Optional[str]](self._process_env)  # Make a copy
        if extra_env_vars:
            result.update(**extra_env_vars)
        for identifier in self._environments_entered:
            if identifier in self._created_env_vars:
                self._created_env_vars[identifier].apply_to_environment(result)
        return result



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import stat
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from shutil import chown
from tempfile import mkstemp
from typing import Any, Generator, Optional, cast

from openjd.model import SymbolTable, FormatStringError
from openjd.model.v2023_09 import EmbeddedFileText as EmbeddedFileText_2023_09
from openjd.model.v2023_09 import (
    ValueReferenceConstants as ValueReferenceConstants_2023_09,
)
from ._logging import LoggerAdapter, LogExtraInfo, LogContent
from ._session_user import PosixSessionUser, SessionUser, WindowsSessionUser
from ._types import EmbeddedFilesListType, EmbeddedFileType

from ._windows_permission_helper import WindowsPermissionHelper
from ._os_checker import is_windows

if is_windows():
    from ._win32._helpers import get_process_user  # type: ignore

__all__ = ("EmbeddedFilesScope", "EmbeddedFiles")


@contextmanager
def _open_context(*args: Any, **kwargs: Any) -> Generator[int, None, None]:
    fd = os.open(*args, **kwargs)
    try:
        yield fd
    finally:
        os.close(fd)


def write_file_for_user(
    filename: Path, data: str, user: Optional[SessionUser], additional_permissions: int = 0
) -> None:
    # File should only be r/w by the owner, by default

    # flags:
    #  O_WRONLY - open for writing
    #  O_CREAT - create if it does not exist
    #  O_TRUNC - truncate the file. If we overwrite an existing file, then we
    #            need to clear its contents.
    #  O_EXCL (intentionally not present) - fail if file exists
    #    - We exclude this 'cause we expect to be writing the same embedded file
    #      into the same location repeatedly with different contents as we run
    #      multiple Tasks in the same Session.
    flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
    # mode:
    #  S_IRUSR - Read by owner
    #  S_IWUSR - Write by owner
    mode = stat.S_IRUSR | stat.S_IWUSR | (additional_permissions & stat.S_IRWXU)
    with _open_context(filename, flags, mode=mode) as fd:
        os.write(fd, data.encode("utf-8"))

    if os.name == "posix":
        if user is not None:
            user = cast(PosixSessionUser, user)
            # Set the group of the file
            chown(filename, group=user.group)
            # Update the permissions to include the group after the group is changed
            # Note: Only after changing group for security in case the group-ownership
            # change fails.
            mode |= stat.S_IRGRP | stat.S_IWGRP | (additional_permissions & stat.S_IRWXG)

        # The file may have already existed before calling this function (e.g. created by mkstemp)
        # so unconditionally set the file permissions to ensure that additional_permissions are set.
        os.chmod(filename, mode=mode)

    elif os.name == "nt":
        if user is not None:
            user = cast(WindowsSessionUser, user)
            process_user = get_process_user()
            WindowsPermissionHelper.set_permissions(
                str(filename),
                principals_full_control=[process_user],
                principals_modify_access=[user.user],
            )


class EmbeddedFilesScope(Enum):
    """What scope of Script a given set of files is for.
    This dictates what prefix is used in format string variables
    """

    STEP = "step"
    ENV = "environment"


@dataclass(frozen=True)
class _FileRecord:
    symbol: str
    filename: Path
    file: EmbeddedFileType


# Note: "EmbeddedFiles" is currently "Attachments" in the Open Job Description template, but that
# will be changing to "EmbeddedFiles" to eliminate potential confusion with job bundle's
# "attachments"
class EmbeddedFiles:
    """Functionality for materializing a Script's Embedded Files to disk, and
    adding their values to a SymbolTable for use in the Script's Actions.
    """

    def __init__(
        self,
        *,
        logger: LoggerAdapter,
        scope: EmbeddedFilesScope,
        session_files_directory: Path,
        user: Optional[SessionUser] = None,
    ) -> None:
        """
        Arguments:
            logger (LoggerAdapter): Logger to send any logging messages to (e.g. errors).
            scope (EmbeddedFilesKind): The scope of the embedded files (used to determine
                value reference prefix in Format Strings).
            session_files_directory (Path): Directory within which to materialize the files to disk.
            user (Optional[SessionUser]): A group that will own the created files.
                The group rw bits will be set on the file if this option is supplied.
                Defaults to current user.
        """
        self._logger = logger
        self._scope = scope
        self._target_directory = session_files_directory
        self._user = user

    def materialize(self, files: EmbeddedFilesListType, symtab: SymbolTable) -> None:
        if self._scope == EmbeddedFilesScope.ENV:
            self._logger.info("Writing embedded files for Environment to disk.")
        else:
            self._logger.info("Writing embedded files for Task to disk.")

        try:
            records = list[_FileRecord]()
            # Generate the symbol table values and filenames
            for file in files:
                # Raises: OSError
                symbol, filename = self._get_symtab_entry(file)
                records.append(_FileRecord(symbol=symbol, filename=filename, file=file))

            # Add symbols to the symbol table
            for record in records:
                symtab[record.symbol] = str(record.filename)
                self._logger.info(
                    f"Mapping: {record.symbol} -> {record.filename}",
                    extra=LogExtraInfo(
                        openjd_log_content=LogContent.FILE_PATH | LogContent.PARAMETER_INFO
                    ),
                )

            # Write the files to disk.
            for record in records:
                # Raises: OSError
                self._materialize_file(record.filename, record.file, symtab)
        except OSError as err:
            raise RuntimeError(f"Could not write embedded file: {err}")
        except FormatStringError as err:
            # This should *never* happen. All format string contents are
            # checked when building the Job Template model. If we get here,
            # then something is broken with our model validation.
            raise RuntimeError(f"Error resolving format string: {str(err)}")

    def _find_value_prefix(self, file: EmbeddedFileType) -> str:
        """Figure out what prefix to use when referencing the file in format strings.
        We figure this out based on the model that `file` comes from and
        self._scope.
        """
        # When adding a new schema, start this method with a check for which
        # model 'file' belongs to -- that'll tell us the schema version.
        assert isinstance(file, EmbeddedFileText_2023_09)

        if self._scope == EmbeddedFilesScope.ENV:
            return ValueReferenceConstants_2023_09.ENV_FILE_PREFIX.value
        else:
            return ValueReferenceConstants_2023_09.TASK_FILE_PREFIX.value

    def _get_symtab_entry(self, file: EmbeddedFileType) -> tuple[str, Path]:
        """Figure out the entry to add to the symbol table for the given
        file. The value of the symbol table entry is the absolute filename
        of the file that we manifest on disk.

        Note: If a random filename is generated, then this does create
           the file as empty to reserve the filename on the filesystem.

        Returns:
            (symbol, value):
                symbol - The symbol to add to the symbol table.
                value - The absolute filename of the file to manifest.
        """

        assert isinstance(file, EmbeddedFileText_2023_09)

        # Figure out what filename to use for the given embedded file.
        # This will either be provided in the given 'file' or we will
        # randomly generate one.
        filename: Path
        if not file.filename:
            # Raises: OSError
            fd, fname = mkstemp(dir=self._target_directory)  # 0o600
            os.close(fd)
            filename = Path(fname)
        else:
            filename = self._target_directory / file.filename

        return (f"{self._find_value_prefix(file)}.{file.name}", filename)

    def _materialize_file(
        self, filename: Path, file: EmbeddedFileType, symtab: SymbolTable
    ) -> None:
        """Materialize/write the file data to disk.
        If self._user is set, then make it r/w by the given group.
        Make the file executable if the file settings indicate that we should.
        """

        assert isinstance(file, EmbeddedFileText_2023_09)

        execute_permissions = 0
        if file.runnable:
            # Allow the owner to execute the file and the group if self._user is set
            execute_permissions |= stat.S_IXUSR | (stat.S_IXGRP if self._user is not None else 0)

        data = file.data.resolve(symtab=symtab)
        # Create the file as r/w owner, and optionally group
        write_file_for_user(filename, data, self._user, additional_permissions=execute_permissions)

        self._logger.info(
            f"Wrote: {file.name} -> {str(filename)}",
            extra=LogExtraInfo(openjd_log_content=LogContent.FILE_PATH),
        )
        self._logger.debug(
            "Contents:\n%s", data, extra=LogExtraInfo(openjd_log_content=LogContent.FILE_CONTENTS)
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from ._logging import LOG, LogContent
from ._path_mapping import PathFormat, PathMappingRule
from ._session import ActionStatus, Session, SessionCallbackType, SessionState
from ._session_user import (
    PosixSessionUser,
    SessionUser,
    WindowsSessionUser,
    BadCredentialsException,
)
from ._types import (
    ActionState,
    EnvironmentIdentifier,
    EnvironmentModel,
    EnvironmentScriptModel,
    StepScriptModel,
)
from ._version import version

__all__ = (
    "ActionState",
    "ActionStatus",
    "EnvironmentIdentifier",
    "EnvironmentModel",
    "EnvironmentScriptModel",
    "LOG",
    "LogContent",
    "PathFormat",
    "PathMappingRule",
    "PosixSessionUser",
    "Session",
    "SessionCallbackType",
    "SessionState",
    "SessionUser",
    "StepScriptModel",
    "WindowsSessionUser",
    "BadCredentialsException",
    "version",
)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
from typing import Optional
from abc import ABC, abstractmethod
from ctypes.wintypes import HANDLE

from ._os_checker import is_posix, is_windows

if is_posix():
    import grp
    import pwd

if is_windows():
    import win32api
    import win32security
    import pywintypes
    import winerror
    from win32con import LOGON32_LOGON_INTERACTIVE, LOGON32_PROVIDER_DEFAULT

    from ._win32._helpers import get_process_user, get_current_process_session_id  # type: ignore

__all__ = (
    "PosixSessionUser",
    "SessionUser",
    "WindowsSessionUser",
    "BadCredentialsException",
)

CURRENT_PROCESS_RUNNING_IN_WINDOWS_SESSION_0: bool
if is_windows():
    CURRENT_PROCESS_RUNNING_IN_WINDOWS_SESSION_0 = 0 == get_current_process_session_id()
else:
    CURRENT_PROCESS_RUNNING_IN_WINDOWS_SESSION_0 = False


class BadCredentialsException(Exception):
    """Exception raised for incorrect username or password."""

    pass


class SessionUser(ABC):
    """Base class for holding information on the specific os-user identity to run
    a Session as.
    """

    user: str
    """
    User name of the identity to run the Session's subprocesses under.
    """

    @staticmethod
    @abstractmethod
    def _get_process_user():
        """
        Returns the user name of the user running the current process.
        """
        pass

    def is_process_user(self) -> bool:
        """
        Returns True if the session user is the user running the current process, else False.

        """
        return self.user == self._get_process_user()


class PosixSessionUser(SessionUser):
    __slots__ = ("user", "group")
    """Specific os-user identity to run a Session as under Linux/macOS."""

    user: str
    """User name of the identity to run the Session's subprocesses under.
    """

    group: str
    """Group name of the identity to run the Session's subprocesses under.
    """

    def __init__(self, user: str, *, group: Optional[str] = None) -> None:
        """
        Arguments:
            user (str): The user
            group (Optional[str]): The group. Defaults to the name of this
                process' effective group.
        """
        if not is_posix():
            raise RuntimeError("Only available on posix systems.")
        self.user = user
        self.group = group if group else grp.getgrgid(os.getegid()).gr_name  # type: ignore

    @staticmethod
    def _get_process_user():
        """
        Returns the user name of the user running the current process.
        """
        return pwd.getpwuid(os.geteuid()).pw_name


class WindowsSessionUser(SessionUser):
    """Specific os-user identity to run a Session as under Windows.

    Note that you must check whether you are running in Windows Session 0 prior to
    creating an instance of this class.
    1. If you're not in Session 0 (i.e. you're in a typical interactive logon via the desktop)
       then you must instantiate this class with a username + password; providing a logon token
       is not allowed at this time.
    2. If you are in Session 0 (i.e. you're running within the context of a Windows Service; this
       includes a logon session obtained by ssh-ing into the host), then you must instantiate this
       class with a username + logon_token; providing a password is not allowed in Session 0. To
       create a logon_token, you will want to look in to the LogonUser family of Win32 system APIs.

    The user provided in this class directly influences the Directory ACL of the Session Working
    Directory that is created. The created directory:
    1. Has Full Control by the owner of the calling process; and
    2. Has Modify access by the provided user.
    The Session working directory will also be set so that all child directories and files
    inherit these permissions.
    """

    __slots__ = ("user", "password", "logon_token")

    user: str
    """
    User name of the identity to run the Session's subprocesses under.
    This can be either a plain username for a local user or a domain username in down-level logon form
    ex: localUser, domain\\domainUser
    """

    password: Optional[str]
    """
    Password of the identity to run the Session's subprocess(es) under.
    Mutually exclusive with: logon_token
    """

    logon_token: Optional[HANDLE]
    """
    A logon token to use to run the Session's subprocess(es) under.
    Mutually exclusive with: password
    """

    def __init__(
        self,
        user: str,
        *,
        password: Optional[str] = None,
        logon_token: Optional[HANDLE] = None,
    ) -> None:
        """
        Arguments:
            user (str):
                User name of the identity to run the Session's subprocesses under.
                This can be either a plain username for a local user, a domain username in down-level logon form,
                or a domain's UPN.
                ex: localUser, domain\\domainUser, domainUser@domain.com

            password (Optional[str]):
                Password of the identity to run the Session's subprocess under. This argument is mutually-exclusive with the
                "logon_token" argument.

            logon_token (Optional[ctypes.wintypes.HANDLE]):
                Windows logon handle for the target user. This argument is mutually-exclusive with the
                "password" argument.
        """
        if not is_windows():
            raise RuntimeError("Only available on Windows systems.")

        if password and logon_token:
            raise ValueError('The "password" and "logon_token" arguments are mutually exclusive')

        self.user = user

        # Note: We allow user to be the process user to support the case of being able to supply
        # the group that the process will run under; differing from the user's default group.
        if self.is_process_user():
            if password is not None:
                raise RuntimeError("User is the process owner. Do not provide a password.")
            if logon_token is not None:
                raise RuntimeError("User is the process owner. Do not provide a logon token.")
        else:
            # Note: "" is allowed as that may actually be the password for the user.
            if password is None and logon_token is None:
                raise RuntimeError(
                    "Must supply a password or logon token. User is not the process owner."
                )
            if password is not None:
                if CURRENT_PROCESS_RUNNING_IN_WINDOWS_SESSION_0:
                    raise RuntimeError(
                        (
                            "Must supply a logon_token rather than a password. "
                            "Passwords are not supported when running in Windows Session 0."
                        )
                    )
                self._validate_username_password(user, password)
                self.password = password
                self.logon_token = None
            else:
                self.password = None
                self.logon_token = logon_token

    @staticmethod
    def _get_process_user():
        return get_process_user()

    @staticmethod
    def _validate_username_password(user_name: str, password: str) -> Optional[bool]:
        """
        Validates the username and password against Windows authentication.

        Args:
            user_name (str): The username to be validated.
            domain_name (str): The domain where the user exists. None means current domain.
            password (str): The password to be validated.

        Returns:
            Optional[bool]: True if the credentials are valid

        Raises:
            BadCredentialsException: If the username or password is incorrect.
        """
        if "\\" in user_name:
            domain, username = user_name.split("\\")
        else:
            domain, username = None, user_name
        try:
            handle = win32security.LogonUser(
                username,
                domain,
                password,
                LOGON32_LOGON_INTERACTIVE,
                LOGON32_PROVIDER_DEFAULT,
            )
            win32api.CloseHandle(handle)
            return True
        except pywintypes.error as e:
            if e.winerror == winerror.ERROR_LOGON_FAILURE:
                raise BadCredentialsException("The username or password is incorrect.")
            raise
